import math

import cv2
import numpy as np
import torch
import torch.nn.functional as F


def gaussian_multiple_channels(num_channels, sigma):
    r = 2 * sigma
    size = 2 * r + 1
    size = int(math.ceil(size))
    x = torch.arange(0, size, 1, dtype=torch.float)
    y = x.unsqueeze(1)
    x0 = y0 = r

    gaussian = torch.exp(-1 * (((x - x0) ** 2 + (y - y0) ** 2) / (2 * (sigma**2)))) / (
        (2 * math.pi * (sigma**2)) ** 0.5
    )
    gaussian = gaussian.to(dtype=torch.float32)

    weights = torch.zeros((num_channels, num_channels, size, size), dtype=torch.float32)
    for i in range(num_channels):
        weights[i, i, :, :] = gaussian

    return weights


def apply_nms(score_map, size):
    from scipy.ndimage import maximum_filter

    score_map = score_map * (score_map == maximum_filter(score_map, footprint=np.ones((size, size))))
    return score_map


def remove_borders(images, borders):
    ## input [B,C,H,W]
    shape = images.shape

    if len(shape) == 4:
        for batch_id in range(shape[0]):
            images[batch_id, :, 0:borders, :] = 0
            images[batch_id, :, :, 0:borders] = 0
            images[batch_id, :, shape[2] - borders : shape[2], :] = 0
            images[batch_id, :, :, shape[3] - borders : shape[3]] = 0
    elif len(shape) == 2:
        images[0:borders, :] = 0
        images[:, 0:borders] = 0
        images[shape[0] - borders : shape[0], :] = 0
        images[:, shape[1] - borders : shape[1]] = 0
    else:
        print("Not implemented")
        exit()

    return images


def getAff(x, y, H):
    h11 = H[0, 0]
    h12 = H[0, 1]
    h13 = H[0, 2]
    h21 = H[1, 0]
    h22 = H[1, 1]
    h23 = H[1, 2]
    h31 = H[2, 0]
    h32 = H[2, 1]
    h33 = H[2, 2]
    fxdx = h11 / (h31 * x + h32 * y + h33) - (h11 * x + h12 * y + h13) * h31 / (h31 * x + h32 * y + h33) ** 2
    fxdy = h12 / (h31 * x + h32 * y + h33) - (h11 * x + h12 * y + h13) * h32 / (h31 * x + h32 * y + h33) ** 2

    fydx = h21 / (h31 * x + h32 * y + h33) - (h21 * x + h22 * y + h23) * h31 / (h31 * x + h32 * y + h33) ** 2
    fydy = h22 / (h31 * x + h32 * y + h33) - (h21 * x + h22 * y + h23) * h32 / (h31 * x + h32 * y + h33) ** 2

    Aff = [[fxdx, fxdy], [fydx, fydy]]

    return np.asarray(Aff)


def apply_homography_to_points(points, h):
    new_points = []

    for point in points:
        new_point = h.dot([point[0], point[1], 1.0])

        tmp = point[2] ** 2 + np.finfo(np.float32).eps

        Mi1 = [[1 / tmp, 0], [0, 1 / tmp]]
        Mi1_inv = np.linalg.inv(Mi1)
        Aff = getAff(point[0], point[1], h)

        BMB = np.linalg.inv(np.dot(Aff, np.dot(Mi1_inv, np.matrix.transpose(Aff))))

        [e, _] = np.linalg.eig(BMB)
        new_radious = 1 / ((e[0] * e[1]) ** 0.5) ** 0.5

        new_point = [
            new_point[0] / new_point[2],
            new_point[1] / new_point[2],
            new_radious,
            point[3],
        ]
        new_points.append(new_point)

    return np.asarray(new_points)


def find_index_higher_scores(map, num_points=1000, threshold=-1):
    # Best n points
    if threshold == -1:
        flatten = map.flatten()
        order_array = np.sort(flatten)

        order_array = np.flip(order_array, axis=0)

        if order_array.shape[0] < num_points:
            num_points = order_array.shape[0]

        threshold = order_array[num_points - 1]

        if threshold <= 0.0:
            ### This is the problem case which derive smaller number of keypoints than the argument "num_points".
            indexes = np.argwhere(order_array > 0.0)

            if len(indexes) == 0:
                threshold = 0.0
            else:
                threshold = order_array[indexes[len(indexes) - 1]]

    indexes = np.argwhere(map >= threshold)

    return indexes[:num_points]


def get_point_coordinates(map, scale_value=1.0, num_points=1000, threshold=-1, order_coord="xysr"):
    ## input numpy array score map : [H, W]
    indexes = find_index_higher_scores(map, num_points=num_points, threshold=threshold)
    new_indexes = []
    for ind in indexes:
        scores = map[ind[0], ind[1]]
        if order_coord == "xysr":
            tmp = [ind[1], ind[0], scale_value, scores]
        elif order_coord == "yxsr":
            tmp = [ind[0], ind[1], scale_value, scores]

        new_indexes.append(tmp)

    indexes = np.asarray(new_indexes)

    return np.asarray(indexes)


class REKD(torch.nn.Module):
    def __init__(self, args, device):
        super(REKD, self).__init__()
        from e2cnn import gspaces, nn

        self.pyramid_levels = 3
        self.factor_scaling = args.factor_scaling_pyramid

        # Smooth Gausian Filter
        num_channels = 1  ## gray scale image
        self.gaussian_avg = gaussian_multiple_channels(num_channels, 1.5)

        r2_act = gspaces.Rot2dOnR2(N=args.group_size)

        self.feat_type_in = nn.FieldType(
            r2_act, num_channels * [r2_act.trivial_repr]
        )  ## input 1 channels (gray scale image)

        feat_type_out1 = nn.FieldType(r2_act, args.dim_first * [r2_act.regular_repr])
        feat_type_out2 = nn.FieldType(r2_act, args.dim_second * [r2_act.regular_repr])
        feat_type_out3 = nn.FieldType(r2_act, args.dim_third * [r2_act.regular_repr])

        feat_type_ori_est = nn.FieldType(r2_act, [r2_act.regular_repr])

        self.block1 = nn.SequentialModule(
            nn.R2Conv(self.feat_type_in, feat_type_out1, kernel_size=5, padding=2, bias=False),
            nn.InnerBatchNorm(feat_type_out1),
            nn.ReLU(feat_type_out1, inplace=True),
        )
        self.block2 = nn.SequentialModule(
            nn.R2Conv(feat_type_out1, feat_type_out2, kernel_size=5, padding=2, bias=False),
            nn.InnerBatchNorm(feat_type_out2),
            nn.ReLU(feat_type_out2, inplace=True),
        )
        self.block3 = nn.SequentialModule(
            nn.R2Conv(feat_type_out2, feat_type_out3, kernel_size=5, padding=2, bias=False),
            nn.InnerBatchNorm(feat_type_out3),
            nn.ReLU(feat_type_out3, inplace=True),
        )

        self.ori_learner = nn.SequentialModule(
            nn.R2Conv(
                feat_type_out3, feat_type_ori_est, kernel_size=1, padding=0, bias=False
            )  ## Channel pooling by 8*G -> 1*G conv.
        )
        self.softmax = torch.nn.Softmax(dim=1)

        self.gpool = nn.GroupPooling(feat_type_out3)
        self.last_layer_learner = torch.nn.Sequential(
            torch.nn.BatchNorm2d(num_features=args.dim_third * self.pyramid_levels),
            torch.nn.Conv2d(
                in_channels=args.dim_third * self.pyramid_levels,
                out_channels=1,
                kernel_size=1,
                bias=True,
            ),
            torch.nn.ReLU(inplace=True),  ## clamp to make the scores positive values.
        )

        self.dim_third = args.dim_third
        self.group_size = args.group_size
        self.exported = False

    def export(self):
        from e2cnn import nn

        for name, module in dict(self.named_modules()).copy().items():
            if isinstance(module, nn.EquivariantModule):
                # print(name, "--->", module)
                module = module.export()
                setattr(self, name, module)

        self.exported = True

    def forward(self, input_data):
        features_key, features_o = self.compute_features(input_data)

        return features_key, features_o

    def compute_features(self, input_data):
        B, _, H, W = input_data.shape

        for idx_level in range(self.pyramid_levels):
            with torch.no_grad():
                input_data_resized = self._resize_input_image(input_data, idx_level, H, W)

            if H > 2500 or W > 2500:
                features_t, features_o = self._forwarding_networks_divide_grid(input_data_resized)
            else:
                features_t, features_o = self._forwarding_networks(input_data_resized)

            features_t = F.interpolate(features_t, size=(H, W), align_corners=True, mode="bilinear")
            features_o = F.interpolate(features_o, size=(H, W), align_corners=True, mode="bilinear")

            if idx_level == 0:
                features_key = features_t
                features_ori = features_o
            else:
                features_key = torch.cat([features_key, features_t], axis=1)
                features_ori = torch.add(features_ori, features_o)

        features_key = self.last_layer_learner(features_key)
        features_ori = self.softmax(features_ori)

        return features_key, features_ori

    def _forwarding_networks(self, input_data_resized):
        from e2cnn import nn

        # wrap the input tensor in a GeometricTensor (associate it with the input type)
        features_t = (
            nn.GeometricTensor(input_data_resized, self.feat_type_in) if not self.exported else input_data_resized
        )

        ## Geometric tensor feed forwarding
        features_t = self.block1(features_t)
        features_t = self.block2(features_t)
        features_t = self.block3(features_t)

        ## orientation pooling
        features_o = self.ori_learner(features_t)  ## self.cpool
        features_o = features_o.tensor if not self.exported else features_o

        ## keypoint pooling
        features_t = self.gpool(features_t)
        features_t = features_t.tensor if not self.exported else features_t

        return features_t, features_o

    def _forwarding_networks_divide_grid(self, input_data_resized):
        device = input_data_resized.device
        
        ## for inference time high resolution image. # spatial grid 4
        B, _, H_resized, W_resized = input_data_resized.shape
        features_t = torch.zeros(B, self.dim_third, H_resized, W_resized).to(device)
        features_o = torch.zeros(B, self.group_size, H_resized, W_resized).to(device)
        h_divide = 2
        w_divide = 2
        for idx in range(h_divide):
            for jdx in range(w_divide):
                ## compute the start and end spatial index
                h_start = H_resized // h_divide * idx
                w_start = W_resized // w_divide * jdx
                h_end = H_resized // h_divide * (idx + 1)
                w_end = W_resized // w_divide * (jdx + 1)
                ## crop the input image
                input_data_divided = input_data_resized[:, :, h_start:h_end, w_start:w_end]
                features_t_temp, features_o_temp = self._forwarding_networks(input_data_divided)
                ## take into the values.
                features_t[:, :, h_start:h_end, w_start:w_end] = features_t_temp
                features_o[:, :, h_start:h_end, w_start:w_end] = features_o_temp

        return features_t, features_o

    def _resize_input_image(self, input_data, idx_level, H, W):
        if idx_level == 0:
            input_data_smooth = input_data
        else:
            ## (7,7) size gaussian kernel.
            input_data_smooth = F.conv2d(input_data, self.gaussian_avg.to(input_data.device), padding=[3, 3])

        target_resize = (
            int(H / (self.factor_scaling**idx_level)),
            int(W / (self.factor_scaling**idx_level)),
        )

        input_data_resized = F.interpolate(input_data_smooth, size=target_resize, align_corners=True, mode="bilinear")

        input_data_resized = self.local_norm_image(input_data_resized)

        return input_data_resized

    def local_norm_image(self, x, k_size=65, eps=1e-10):
        pad = int(k_size / 2)

        x_pad = F.pad(x, (pad, pad, pad, pad), mode="reflect")
        x_mean = F.avg_pool2d(x_pad, kernel_size=[k_size, k_size], stride=[1, 1], padding=0)  ## padding='valid'==0
        x2_mean = F.avg_pool2d(
            torch.pow(x_pad, 2.0),
            kernel_size=[k_size, k_size],
            stride=[1, 1],
            padding=0,
        )

        x_std = torch.sqrt(torch.abs(x2_mean - x_mean * x_mean)) + eps
        x_norm = (x - x_mean) / (1.0 + x_std)

        return x_norm


def load_detector(args, device):
    args.group_size, args.dim_first, args.dim_second, args.dim_third = model_parsing(args)
    model1 = REKD(args, device)
    model1.load_state_dict(torch.load(args.weights, weights_only=True, map_location=device))
    model1.export()
    model1.eval()
    model1.to(device)  ## use GPU

    return model1


## Load our model
def model_parsing(args):
    # get the dir/model.pt
    base_path = "/".join(args.weights.split("/")[-2:])
    
    group_size = base_path.split("_group")[1].split("_")[0]
    dim_first = base_path.split("_f")[1].split("_")[0]
    dim_second = base_path.split("_s")[1].split("_")[0]
    dim_third = base_path.split("_t")[1].split(".log")[0]
    
    return int(group_size), int(dim_first), int(dim_second), int(dim_third)


def upsample_pyramid(image, upsampled_levels, scale_factor_levels):
    ## image np.array([C, H, W]), upsampled_levels int
    up_pyramid = []
    for j in range(upsampled_levels):
        factor = scale_factor_levels ** (upsampled_levels - j)
        up_image = cv2.resize(
            image.transpose(1, 2, 0),
            dsize=(0, 0),
            fx=factor,
            fy=factor,
            interpolation=cv2.INTER_LINEAR,
        )
        up_pyramid.append(up_image[np.newaxis])

    return up_pyramid



import torch
from omegaconf import OmegaConf

from easy_local_features.submodules.git_alike.alike import ALike

from ..matching.nearest_neighbor import NearestNeighborMatcher
from ..utils import ops
from ..utils.download import downloadModel
from .basemodel import BaseExtractor, MethodType
from typing import TypedDict, Optional


class ALIKEConfig(TypedDict):
    model_name: str
    top_k: int
    scores_th: float
    n_limit: int
    sub_pixel: bool
    model_path: Optional[str]


models = {
    "alike-t": "https://github.com/Shiaoming/ALIKE/raw/main/models/alike-t.pth",
    "alike-s": "https://github.com/Shiaoming/ALIKE/raw/main/models/alike-s.pth",
    "alike-n": "https://github.com/Shiaoming/ALIKE/raw/main/models/alike-n.pth",
    "alike-l": "https://github.com/Shiaoming/ALIKE/raw/main/models/alike-l.pth",
}

configs = {
    "alike-t": {"c1": 8, "c2": 16, "c3": 32, "c4": 64, "dim": 64, "single_head": True, "radius": 2, "model_path": ""},
    "alike-s": {"c1": 8, "c2": 16, "c3": 48, "c4": 96, "dim": 96, "single_head": True, "radius": 2, "model_path": ""},
    "alike-n": {
        "c1": 16,
        "c2": 32,
        "c3": 64,
        "c4": 128,
        "dim": 128,
        "single_head": True,
        "radius": 2,
        "model_path": "",
    },
    "alike-l": {
        "c1": 32,
        "c2": 64,
        "c3": 128,
        "c4": 128,
        "dim": 128,
        "single_head": False,
        "radius": 2,
        "model_path": "",
    },
}


class ALIKE_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DETECT_DESCRIBE
    """ALIKE baseline implementation.
    model_name: str = 'alike-t' | 'alike-s' | 'alike-n' | 'alike-l'
    top_k: int = -1. Detect top K keypoints. -1 for threshold based mode, >0 for top K mode. (default: -1)
    scores_th: float = 0.2. Detector score threshold (default: 0.2).
    n_limit: int = 5000. Maximum number of keypoints to be detected (default: 5000).
    no_sub_pixel: bool = False. Do not detect sub-pixel keypoints (default: False).
    device: int = -1. Device to run the model on. -1 for CPU, >=0 for GPU. (default: -1)
    """

    default_conf: ALIKEConfig = {
        "model_name": "alike-t",
        "top_k": -1,
        "scores_th": 0.2,
        "n_limit": 2048,
        "sub_pixel": True,
        "model_path": None,
    }

    def __init__(self, conf: ALIKEConfig = {}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        model_name = conf.model_name
        top_k = conf.top_k
        scores_th = conf.scores_th
        n_limit = conf.n_limit
        sub_pixel = conf.sub_pixel
        model_path = conf.model_path
        self.DEV = torch.device("cpu")

        if model_name not in models:
            raise ValueError(f"Model name {model_name} not found in {models.keys()}")

        if model_path is None:
            url = models[model_name]
            model_path = downloadModel("alike", model_name, url)

        config = configs[model_name]
        config["model_path"] = model_path

        self.model = ALike(**config, top_k=top_k, scores_th=scores_th, n_limit=n_limit).to(self.DEV)

        self.sub_pixel = sub_pixel
        self.matcher = NearestNeighborMatcher()

    def detectAndCompute(self, image, return_dict=False):
        image = ops.prepareImage(image)
        image = image.to(self.DEV)
        pred = self.model(image, sub_pixel=self.sub_pixel)
        keypoints = pred["keypoints"]  # (N, 2)
        descriptors = pred["descriptors"]  # (N, 64)
        scores = pred["scores"]  # (N,)

        keypoints = keypoints.unsqueeze(0)
        descriptors = descriptors.unsqueeze(0)
        scores = scores.unsqueeze(0)

        if return_dict:
            return {"keypoints": keypoints, "descriptors": descriptors, "scores": scores}

        return keypoints, descriptors

    def detect(self, img):
        return self.detectAndCompute(img)[0]

    def compute(self, image, keypoints):
        raise NotImplementedError("ALIKE_baseline.compute is not implemented; use detectAndCompute instead.")

    def to(self, device):
        self.model.to(device)
        self.DEV = device

    @property
    def has_detector(self):
        return True



from e2cnn.nn import *
from e2cnn.gspaces import *

import torch
import torch.nn as nn
import numpy as np

class E2SFCNN_QUOT(torch.nn.Module):
    
    def __init__(self, n_channels, n_classes,
                 N=None,
                 restrict: int = -1,
                 fix_param: bool = False,
                 fco: float = 0.8,
                 p_drop_fully: float = 0.3,
                 J: int = 0,
                 sigma: float = 0.6,
                 sgsize: int = None,
                 flip: bool = True,
                 ):
        
        super(E2SFCNN_QUOT, self).__init__()

        if N is None:
            N = 16
        
        assert N > 1
        
        self.N = N
        self.n_channels = n_channels
        self.n_classes = n_classes

        # build the group O(2) or D_N depending on the number N of rotations specified
        if N > 1:
            self.gspace = FlipRot2dOnR2(N)
        elif N == 1:
            self.gspace = Flip2dOnR2()
        else:
            raise ValueError(N)

        # if flips are not required, immediately restrict to the SO(2) or C_N subgroup
        if not flip:
            if N != 1:
                sg = (None, N)
            else:
                sg = 1
            self.gspace, _, _ = self.gspace.restrict(sg)

        # id of the subgroup if group restriction is applied through the network
        if sgsize is not None:
            self.sgid = sgsize
        else:
            self.sgid = N

        if flip and N != 1:
            self.sgid = (None, self.sgid)

        if fco is not None and fco > 0.:
            fco *= np.pi
        self.frequencies_cutoff = fco
        self.sigma = sigma
        self.J = J
        self.fix_param = fix_param
        self.restrict = restrict
        
        eq_layers = []

        self.LAYER = 0
        
        if restrict == self.LAYER:
            gspace, _, _ = self.gspace.restrict(self.sgid)
        
        r1 = FieldType(gspace, [gspace.trivial_repr] * n_channels)
        
        # 28 px
        # Convolutional Layer 1

        self.LAYER += 1
        eq_layers += self.build_layer_quotient(r1, 24, 9, 0, None)
        
        # Convolutional Layer 2
        self.LAYER += 1
        eq_layers += self.build_layer_quotient(eq_layers[-1].out_type, 32, 7, 3, 2)
        
        # 14 px
        # Convolutional Layer 3
        self.LAYER += 1
        eq_layers += self.build_layer_quotient(eq_layers[-1].out_type, 36, 7, 3, None)

        # Convolutional Layer 4
        self.LAYER += 1
        eq_layers += self.build_layer_quotient(eq_layers[-1].out_type, 36, 7, 3, 2)

        # 7 px
        # Convolutional Layer 5
        self.LAYER += 1
        eq_layers += self.build_layer_quotient(eq_layers[-1].out_type, 64, 7, 3)

        # Convolutional Layer 6
        self.LAYER += 1
        eq_layers += self.build_layer_quotient(eq_layers[-1].out_type, 96, 5, 0, None, True)

        # Adaptive Pooling
        mpl = PointwiseAdaptiveMaxPool(eq_layers[-1].out_type, 1)
        eq_layers.append(mpl)

        # 1 px
        
        # c = 96
        c = eq_layers[-1].out_type.size
        
        self.in_repr = eq_layers[0].in_type
        self.eq_layers = torch.nn.ModuleList(eq_layers)

        # Fully Connected
        self.fully_net = nn.Sequential(
            nn.Dropout(p=p_drop_fully),
            nn.Linear(c, 96),
            nn.BatchNorm1d(96),
            nn.ELU(inplace=True),
    
            nn.Dropout(p=p_drop_fully),
            nn.Linear(96, 96),
            nn.BatchNorm1d(96),
            nn.ELU(inplace=True),
            
            nn.Dropout(p=p_drop_fully),
            nn.Linear(96, n_classes),
        )
    
    def forward(self, input: torch.tensor):
        x = GeometricTensor(input, self.in_repr)
        
        for layer in self.eq_layers:
            x = layer(x)
        
        x = self.fully_net(x.tensor.reshape(x.tensor.shape[0], -1))
        
        return x
        
    def features(self, input):
        x = GeometricTensor(input, self.in_repr)

        # layer_taken = [2, 4, 6, 11, 13, 16, 18, 20] ## all changed (channel expansion, resolution drop)
        layer_taken = [4, 11, 18]   ## before resolution drop

        out = []
        for idx, layer in enumerate(self.eq_layers):
            x = layer(x)
            if idx in layer_taken:
                out.append(x)

        return out[0], out[1], out[2]

    def build_quotient_feature_type(self, gspace):
    
        assert gspace.fibergroup.order() > 0
        if isinstance(gspace, FlipRot2dOnR2):
            n = int(gspace.fibergroup.order() / 2)
            repr = [gspace.regular_repr] * 5
            for i in [0, round(n / 4), round(n / 2)]:
                repr += [gspace.quotient_repr((int(i), 1))] * 2
            repr += [gspace.quotient_repr((None, int(n / 2)))] * 2
            repr += [gspace.trivial_repr] * int(gspace.fibergroup.order() / 4)
        elif isinstance(gspace, Rot2dOnR2):
            n = gspace.fibergroup.order()
            repr = [gspace.regular_repr] * 5
            repr += [gspace.quotient_repr(int(round(n / 2)))] * 2
            repr += [gspace.quotient_repr(int(round(n / 4)))] * 2
            repr += [gspace.trivial_repr] * int(gspace.fibergroup.order() / 4)
        else:
            repr = [gspace.regular_repr]
        
        return repr

    def build_layer_quotient(self, r1: FieldType, C: int, s: int, padding: int = 0, pooling: int = None,
                             orientation_pooling: bool = False):
    
        gspace = r1.gspace
    
        if self.fix_param and not orientation_pooling and self.LAYER > 1:
            # to keep number of parameters more or less constant when changing groups
            # (more precisely, we try to keep them close to the number of parameters in the original SFCNN)
            t = gspace.fibergroup.order() / 16
            C = C / np.sqrt(t)
    
        layers = []
    
        repr = self.build_quotient_feature_type(gspace)
        
        C /= sum([r.size for r in repr]) / gspace.fibergroup.order()
    
        C = int(round(C))
    
        r2 = FieldType(gspace, repr * C).sorted()
    
        cl = R2Conv(r1, r2, s,
                    frequencies_cutoff=self.frequencies_cutoff,
                    padding=padding,
                    sigma=self.sigma,
                    maximum_offset=self.J)
        layers.append(cl)
    
        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, self.sgid))
            layers.append(DisentangleModule(layers[-1].out_type))
    
        if orientation_pooling:
            pl = GroupPooling(layers[-1].out_type)
            layers.append(pl)
    
        bn = InnerBatchNorm(layers[-1].out_type)
        layers.append(bn)
        nnl = ELU(layers[-1].out_type, inplace=True)
        layers.append(nnl)
    
        if pooling is not None:
            pl = PointwiseMaxPool(layers[-1].out_type, pooling)
            layers.append(pl)
    
        return layers




import numpy as np
import os
import cv2
from tqdm import tqdm
import json
import multiprocessing as mp
import torch

from easy_local_features.utils import vis, io
import argparse

try:
    import poselib
except:
    print("Please install poselib library with `pip install poselib`")

def intrinsics_to_camera(K):
    px, py = K[0, 2], K[1, 2]
    fx, fy = K[0, 0], K[1, 1]
    return {
        "model": "PINHOLE",
        "width": int(2 * px),
        "height": int(2 * py),
        "params": [fx, fy, px, py],
    }

# 
def plot_matches_parallel(args_plot):
    pair_idx, pair, mkpts0, mkpts1, out_folder = args_plot
    out_path = os.path.join(out_folder, f'{pair_idx}.png')
    image0 = cv2.imread(pair['image0'])
    image1 = cv2.imread(pair['image1'])
    if isinstance(mkpts0, torch.Tensor):
        mkpts0 = mkpts0.cpu().numpy()
        mkpts1 = mkpts1.cpu().numpy()
        
    if len(mkpts0) == 0 or len(mkpts1) == 0:
        fig, ax = vis.plot_pair(image0, image1)
        vis.save(out_path)
        vis.plt.close('all')
        return True

    (pose,details) = poselib.estimate_relative_pose(
            mkpts0.tolist(), 
            mkpts1.tolist(),
            intrinsics_to_camera(pair['K0']),
            intrinsics_to_camera(pair['K1']),
            ransac_opt={
                'max_iterations': 10000, # default 100000
                'success_prob': 0.99999, # default 0.99999
                'max_epipolar_error': 6.0,
            })
    inliers = np.array(details['inliers'])

    fig, ax = vis.plot_pair(image0, image1)
    vis.plot_matches(mkpts0[~inliers], mkpts1[~inliers], color='r')
    vis.plot_matches(mkpts0[inliers], mkpts1[inliers], color='g')
    vis.add_text(f'Inliers: {inliers.sum()} / {len(inliers)}')
    vis.save(out_path)

    vis.plt.close('all')
    return True

def angle_error_vec(v1, v2):
    n = np.linalg.norm(v1) * np.linalg.norm(v2)
    return np.rad2deg(np.arccos(np.clip(np.dot(v1, v2) / n, -1.0, 1.0)))

def angle_error_mat(R1, R2):
    cos = (np.trace(np.dot(R1.T, R2)) - 1) / 2
    cos = np.clip(cos, -1., 1.)  # numercial errors can make it out of bounds
    return np.rad2deg(np.abs(np.arccos(cos)))

def compute_pose_error(T_0to1, R, t):
    R_gt = T_0to1[:3, :3]
    t_gt = T_0to1[:3, 3]
    error_t = angle_error_vec(t, t_gt)
    error_t = np.minimum(error_t, 180 - error_t)  # ambiguity of E estimation
    error_R = angle_error_mat(R, R_gt)
    return error_t, error_R

def estimate_pose(kpts0, kpts1, K0, K1, thresh, conf=0.99999, pose_estimator='poselib'):
    if len(kpts0) < 5:
        return None
    if pose_estimator == 'poselib':
        import poselib
        (pose,details) = poselib.estimate_relative_pose(
            kpts0.tolist(), 
            kpts1.tolist(),
            intrinsics_to_camera(K0),
            intrinsics_to_camera(K1),
            ransac_opt={
                'max_iterations': 10000, # default 100000
                # 'min_iterations': 1000,
                # 'dyn_num_trials_mult': 3.0,
                'success_prob': conf, # default 0.99999
                # 'max_reproj_error': 12.0,
                'max_epipolar_error': thresh, # default 1.0
                # 'seed': 0,
                # 'progressive_sampling': False,
                # 'max_prosac_iterations': 100000
            },
            bundle_opt={  # all defaults
                # 'max_iterations': 100,
                # 'loss_scale': 1.0,
                # 'loss_type': 'CAUCHY',
                # 'gradient_tol': 1e-10,
                # 'step_tol': 1e-08,
                # 'initial_lambda': 0.001,
                # 'min_lambda': 1e-10,
                # 'max_lambda': 10000000000.0,
                # 'verbose': False
                },
            )
        ret = (pose.R, pose.t, details['inliers'])

    elif pose_estimator == 'opencv':
        f_mean = np.mean([K0[0, 0], K1[1, 1], K0[0, 0], K1[1, 1]])
        norm_thresh = thresh / f_mean

        kpts0 = (kpts0 - K0[[0, 1], [2, 2]][None]) / K0[[0, 1], [0, 1]][None]
        kpts1 = (kpts1 - K1[[0, 1], [2, 2]][None]) / K1[[0, 1], [0, 1]][None]

        E, mask = cv2.findEssentialMat(
            kpts0, kpts1, np.eye(3), threshold=norm_thresh, prob=conf,
            method=cv2.RANSAC)

        assert E is not None

        best_num_inliers = 0
        ret = None
        for _E in np.split(E, len(E) / 3):
            n, R, t, _ = cv2.recoverPose(
                _E, kpts0, kpts1, np.eye(3), 1e9, mask=mask)
            if n > best_num_inliers:
                best_num_inliers = n
                ret = (R, t[:, 0], mask.ravel() > 0)
    else:
        raise NotImplementedError

    return ret

def estimate_pose_parallel(args):
    return estimate_pose(*args)

def pose_auc(errors, thresholds):
    sort_idx = np.argsort(errors)
    errors = np.array(errors.copy())[sort_idx]
    recall = (np.arange(len(errors)) + 1) / len(errors)
    errors = np.r_[0., errors]
    recall = np.r_[0., recall]
    aucs = []
    for t in thresholds:
        last_index = np.searchsorted(errors, t)
        r = np.r_[recall[:last_index], recall[last_index-1]]
        e = np.r_[errors[:last_index], t]
        aucs.append(np.trapz(r, x=e)/t)
    return aucs

def pose_accuracy(errors, thresholds):
    return [np.mean(errors < t) * 100 for t in thresholds]

def resize_long_edge(image, max_size):
    h, w = image.shape[-2:]
    if h > w:
        new_h = max_size
        new_w = int(w * max_size / h)
    else:
        new_w = max_size
        new_h = int(h * max_size / w)
    
    scale = new_h / h
    
    return torch.nn.functional.interpolate(image, size=(new_h, new_w), mode='bilinear', align_corners=False), scale

class PoseEval:
    default_config = {
        'data_path': "",
        'pairs_path': "",
        'pose_estimator': 'poselib', # poselib, opencv
        'cache_images': False,
        'ransac_thresholds': [0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0],
        'pose_thresholds': [5, 10, 20],
        'max_pairs': -1,
        'output': './output/pose_eval/',
        'n_workers': 8,
        'resize': None,
        'detector_only': False,
    }

    def __init__(self, config={}) -> None:
        self.config = {**self.default_config, **config}
        for k,v in self.config.items():
            print(k, v)
        
        self.pairs = self.read_gt()
        print(f'Loaded {len(self.pairs)} pairs')

        os.makedirs(self.config['output'], exist_ok=True)

        if self.config['n_workers'] == -1:
            self.config['n_workers'] = mp.cpu_count()

        self.image_cache = {}
        if self.config['cache_images']:
            self.load_images()

    def load_images(self):
        for pair in tqdm(self.pairs, desc='Caching images'):
            if pair['image0'] not in self.image_cache:
                self.image_cache[pair['image0']] = io.fromPath(pair['image0'])
            if pair['image1'] not in self.image_cache:
                self.image_cache[pair['image1']] = io.fromPath(pair['image1'])

    def read_image(self, path):
        if self.config['cache_images']:
            return self.image_cache[path]
        else:
            return io.fromPath(path)

    def read_gt(self):
        pairs = []
        with open(self.config['pairs_path'], 'r') as f:
            lines = f.readlines()
            for line in lines:
                line = line.replace('\n', '').replace('  ', '')
                all_info = line.split(' ')
                image0 = all_info[0]
                image1 = all_info[1]

                # 2 3x3 matrices and one 4x4 matrix
                if len(all_info) == 38:
                    # scannet format
                    K0 = np.array(all_info[4:13]).astype(float).reshape(3, 3)
                    K1 = np.array(all_info[13:22]).astype(float).reshape(3, 3)
                    T_0to1 = np.array(all_info[22:38]).astype(float).reshape(4, 4)
                    
                    image0 = os.path.join(self.config['data_path'], image0)
                    image1 = os.path.join(self.config['data_path'], image1)
                    depth0 = image0.replace('color', 'depth').replace('jpg', 'png')
                    depth1 = image1.replace('color', 'depth').replace('jpg', 'png')

                    pairs.append({
                        'image0': image0,
                        'image1': image1,
                        'depth0': depth0,
                        'depth1': depth1,
                        'K0': K0,
                        'K1': K1,
                        'T_0to1': T_0to1,
                    })

                # 2 3x3 matrices and one 4x4 matrix
                if len(all_info) == 36:
                    # scannet format
                    K0 = np.array(all_info[2:11]).astype(float).reshape(3, 3)
                    K1 = np.array(all_info[11:20]).astype(float).reshape(3, 3)
                    T_0to1 = np.array(all_info[20:36]).astype(float).reshape(4, 4)
                    
                    image0 = os.path.join(self.config['data_path'], image0)
                    image1 = os.path.join(self.config['data_path'], image1)
                    depth0 = image0.replace('color', 'depth').replace('jpg', 'png')
                    depth1 = image1.replace('color', 'depth').replace('jpg', 'png')

                    pairs.append({
                        'image0': image0,
                        'image1': image1,
                        'depth0': depth0,
                        'depth1': depth1,
                        'K0': K0,
                        'K1': K1,
                        'T_0to1': T_0to1,
                    })
                    
                elif len(all_info) == 32:
                    # megadepth format
                    K0 = np.array(all_info[2:11]).astype(float).reshape(3, 3)
                    K1 = np.array(all_info[11:20]).astype(float).reshape(3, 3)
                    
                    pose_elems = all_info[20:32]
                    R, t = pose_elems[:9], pose_elems[9:12]
                    R = np.array([float(x) for x in R]).reshape(3, 3).astype(np.float32)
                    t = np.array([float(x) for x in t]).astype(np.float32)
                    T_0to1 = np.eye(4)
                    T_0to1[:3, :3] = R
                    T_0to1[:3, 3] = t

                    image0 = os.path.join(self.config['data_path'], image0)
                    image1 = os.path.join(self.config['data_path'], image1)
                    
                    pairs.append({
                        'image0': image0,
                        'image1': image1,
                        'K0': K0,
                        'K1': K1,
                        'T_0to1': T_0to1,
                    })
                else:
                    print(f'Unknown format for pair with {len(all_info)} elements: {line}')
                    raise ValueError(f'Unknown format for pair {line}')

            if self.config['max_pairs'] > 0:
                pairs = pairs[:self.config['max_pairs']]
        return pairs


    def extract_and_save_matches(self, matcher_fn, name='', force=False):
        all_matches = []
        if name == '':
            name = matcher_fn.__name__

        os.makedirs(os.path.join(self.config['output'], name), exist_ok=True)
        fname = os.path.join(self.config['output'], f'{name}/matches.npz')
        
        if not force and os.path.exists(fname):
            return np.load(fname, allow_pickle=True)['all_matches']

        for pair in tqdm(self.pairs, desc='Extracting matches'):
            image0 = self.read_image(pair['image0'])
            image1 = self.read_image(pair['image1'])
            
            if self.config['resize'] is not None:
                image0, scale0 = resize_long_edge(image0, self.config['resize'])
                image1, scale1 = resize_long_edge(image1, self.config['resize'])

            res = matcher_fn(image0, image1)
            mkpts0 = res['mkpts0']
            mkpts1 = res['mkpts1']
            if isinstance(mkpts0, torch.Tensor):
                mkpts0 = mkpts0.cpu().numpy()
                mkpts1 = mkpts1.cpu().numpy()
            
            if self.config['resize'] is not None:
                mkpts0 = mkpts0 / scale0
                mkpts1 = mkpts1 / scale1

            if self.config['detector_only']:
                mkpts0, mkpts1 = self.oracle_matches_scannet(pair, mkpts0, mkpts1)

            all_matches.append({
                'image0': pair['image0'],
                'image1': pair['image1'],
                'mkpts0': mkpts0,
                'mkpts1': mkpts1,
            })
        
        np.savez(fname, all_matches=all_matches)
        
        return all_matches
    
    def oracle_matches_scannet(self, pair, kpts0, kpts1):
        from easy_local_features.utils.ops import warp_kpts
        
        # project points to the other image and get gt matches based on distance
        K0, K1 = pair['K0'], pair['K1']
        T_0to1 = pair['T_0to1']
        depth0 = cv2.imread(pair['depth0'], cv2.IMREAD_ANYDEPTH).astype(np.float32) / 1000
        depth1 = cv2.imread(pair['depth1'], cv2.IMREAD_ANYDEPTH).astype(np.float32) / 1000
        image0 = cv2.imread(pair['image0'])
        image1 = cv2.imread(pair['image1'])
        
        depth0 = cv2.resize(depth0, (image0.shape[1], image0.shape[0]), interpolation=cv2.INTER_NEAREST)
        depth1 = cv2.resize(depth1, (image1.shape[1], image1.shape[0]), interpolation=cv2.INTER_NEAREST)
        
        if not isinstance(kpts0, torch.Tensor):
            kpts0 = torch.tensor(kpts0).float()
            kpts1 = torch.tensor(kpts1).float()
            K0, K1 = torch.tensor(K0).float(), torch.tensor(K1).float()
            T_0to1 = torch.tensor(T_0to1).float()
            depth0 = torch.tensor(depth0).float()
            depth1 = torch.tensor(depth1).float()
        
        valid0, proj1 = warp_kpts(kpts0[None], depth0[None], depth1[None], T_0to1[None], K0[None], K1[None], do_on_cpu=True)
        proj1 = torch.where(valid0[..., None], proj1, torch.tensor(-1.0))[0]
        valid0 = valid0[0]
        
        # cdist between projected points and kpts1
        dists = torch.cdist(proj1, kpts1)
        dist, idx = dists.min(dim=1)
        valid = dist < 1.0
        
        mkpts0 = kpts0[valid]
        mkpts1 = kpts1[idx[valid]]
        
        return mkpts0.numpy(), mkpts1.numpy()
        

    def run_benchmark(self, matcher_fn, name='', force=False):
        if name == '':
            name = matcher_fn.__name__

        all_matches = self.extract_and_save_matches(matcher_fn, name=name, force=force)
        
        # plot matches in parallel
        out_folder = os.path.join(self.config['output'], name, 'matches')
        os.makedirs(out_folder, exist_ok=True)
        plot_args = [ (pair_idx, pair, all_matches[pair_idx]['mkpts0'], all_matches[pair_idx]['mkpts1'], out_folder) for pair_idx, pair in enumerate(self.pairs) ]
        # select one every 10 
        plot_args = plot_args[::10]
        pool = mp.Pool(self.config['n_workers'])
        r = list(tqdm(pool.imap(plot_matches_parallel, plot_args), total=len(plot_args), desc=f'Plotting matches for {name}', leave=False))
        pool.close()
        # for args in tqdm(plot_args, desc=f'Plotting matches for {name}'):
            # vis.plot_matches_parallel(args)
            
        aucs_by_thresh = {}
        accs_by_thresh = {}
        for ransac_thresh in self.config['ransac_thresholds']:

            fname = os.path.join(self.config['output'], f'{name}/{self.config["pose_estimator"]}_{ransac_thresh}.txt')
            # check if exists and has the right number of lines
            if not force and os.path.exists(fname) and len(open(fname, 'r').readlines()) == len(self.pairs) and False:
                errors = []
                with open(fname, 'r') as f:
                    lines = f.readlines()
                    for line in lines:
                        line = line.replace('\n', '')
                        err_t, err_R = line.split(' ')
                        errors.append([float(err_t), float(err_R)])
            # redo the benchmark
            else:
                errors = []
                pairs = self.pairs
                errors_file = open(fname, 'w')

                # do the benchmark in parallel
                if self.config['n_workers'] != 1:
                    

                    pool = mp.Pool(self.config['n_workers'])
                    pool_args = [ (all_matches[pair_idx]['mkpts0'], all_matches[pair_idx]['mkpts1'], pair['K0'], pair['K1'], ransac_thresh, 0.99999, self.config['pose_estimator']) for pair_idx, pair in enumerate(pairs) ]
                    results = list(tqdm(pool.imap(estimate_pose_parallel, pool_args), total=len(pool_args), desc=f'Running benchmark for th={ransac_thresh}', leave=False))
                    pool.close()

                    for pair_idx, ret in enumerate(results):
                        if ret is None:
                            err_t, err_R = np.inf, np.inf
                        else:
                            R, t, inliers = ret
                            pair = pairs[pair_idx]
                            err_t, err_R = compute_pose_error(pair['T_0to1'], R, t)
                        errors_file.write(f'{err_t} {err_R}\n')
                        errors.append([err_t, err_R])
                # do the benchmark in serial
                else:
                    for pair_idx, pair in tqdm(enumerate(pairs), desc=f'Running benchmark for th={ransac_thresh}', leave=False, total=len(pairs)):
                        mkpts0 = all_matches[pair_idx]['mkpts0']
                        mkpts1 = all_matches[pair_idx]['mkpts1']
                        ret = estimate_pose(mkpts0, mkpts1, pair['K0'], pair['K1'], ransac_thresh, 0.99999, self.config['pose_estimator'])
                        if ret is None:
                            err_t, err_R = np.inf, np.inf
                        else:
                            R, t, inliers = ret
                            err_t, err_R = compute_pose_error(pair['T_0to1'], R, t)
                        errors_file.write(f'{err_t} {err_R}\n')
                        errors_file.flush()
                        errors.append([err_t, err_R])

                errors_file.close()

            # compute AUCs
            errors = np.array(errors)
            errors = errors.max(axis=1) 
            aucs = pose_auc(errors, self.config['pose_thresholds'])
            aucs = {k: v*100 for k, v in zip(self.config['pose_thresholds'], aucs)}
            aucs_by_thresh[ransac_thresh] = aucs
            
            accs = pose_accuracy(errors, self.config['pose_thresholds'])
            accs_by_thresh[ransac_thresh] = accs
            

            # dump summary for this method
            summary = {
                'name': name,
                'aucs_by_thresh': aucs_by_thresh,
                'accs_by_thresh': accs_by_thresh,
            } 
            json.dump(summary, open(os.path.join(self.config['output'], f'{name}/{self.config["pose_estimator"]}_summary.json'), 'w'), indent=2)

        return aucs_by_thresh

def parse():
    parser = argparse.ArgumentParser()
    parser.add_argument("--method",  "-m", help="Method to eval")
    parser.add_argument("--benchmark", "-b", help="eval on scannet1500 or megadepth1500")
    return parser.parse_args()


if __name__ == "__main__":
    args = parse()
    
    pose_eval =    this_dir = os.path.dirname(os.path.abspath(__file__))
    from easy_local_features import getExtractor
    method = getExtractor(args.method, {'top_k': 4096})
    megadepth = PoseEval({
        'data_path': '/Users/cadar/Documents/Datasets/scannet_test_1500/',
        'pairs_path': '/Users/cadar/Documents/Datasets/scannet_test_1500/pairs_calibrated.txt',
        'output': './output/pose_eval/scannet/',
        'max_pairs': 100,
        'pose_estimator': 'poselib',
        'ransac_thresholds': [0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0],
        'n_workers': 36,
        # 'resize': -1,
        # 'detector_only': True,
    })
    
    megadepth.run_benchmark(method.match, name=f'xfeat_full', force=True) 
    
    for size in range(400, 2000, 100):
        megadepth = PoseEval({
            'data_path': '/Users/cadar/Documents/Datasets/scannet_test_1500/',
            'pairs_path': '/Users/cadar/Documents/Datasets/scannet_test_1500/pairs_calibrated.txt',
            'output': './output/pose_eval/scannet/',
            'max_pairs': 100,
            'pose_estimator': 'poselib',
            'ransac_thresholds': [0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0],
            'n_workers': 36,
            'resize': size,
            # 'detector_only': True,
        })
        
        megadepth.run_benchmark(method.match, name=f'xfeat_full_{size}', force=True) 



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn as nn
from torch.nn.init import trunc_normal_
from torch.nn.utils import weight_norm


class DINOHead(nn.Module):
    def __init__(
        self,
        in_dim,
        out_dim,
        use_bn=False,
        nlayers=3,
        hidden_dim=2048,
        bottleneck_dim=256,
        mlp_bias=True,
    ):
        super().__init__()
        nlayers = max(nlayers, 1)
        self.mlp = _build_mlp(nlayers, in_dim, bottleneck_dim, hidden_dim=hidden_dim, use_bn=use_bn, bias=mlp_bias)
        self.apply(self._init_weights)
        self.last_layer = weight_norm(nn.Linear(bottleneck_dim, out_dim, bias=False))
        self.last_layer.weight_g.data.fill_(1)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.mlp(x)
        eps = 1e-6 if x.dtype == torch.float16 else 1e-12
        x = nn.functional.normalize(x, dim=-1, p=2, eps=eps)
        x = self.last_layer(x)
        return x


def _build_mlp(nlayers, in_dim, bottleneck_dim, hidden_dim=None, use_bn=False, bias=True):
    if nlayers == 1:
        return nn.Linear(in_dim, bottleneck_dim, bias=bias)
    else:
        layers = [nn.Linear(in_dim, hidden_dim, bias=bias)]
        if use_bn:
            layers.append(nn.BatchNorm1d(hidden_dim))
        layers.append(nn.GELU())
        for _ in range(nlayers - 2):
            layers.append(nn.Linear(hidden_dim, hidden_dim, bias=bias))
            if use_bn:
                layers.append(nn.BatchNorm1d(hidden_dim))
            layers.append(nn.GELU())
        layers.append(nn.Linear(hidden_dim, bottleneck_dim, bias=bias))
        return nn.Sequential(*layers)



import torch
from omegaconf import OmegaConf

from ..matching.nearest_neighbor import NearestNeighborMatcher
from ..submodules.git_sfd2.sfd2 import SFD2
from ..utils import ops
from ..utils.download import downloadModel
from .basemodel import BaseExtractor, MethodType

weights_link = "https://github.com/feixue94/sfd2/raw/dev/weights/20220810_ressegnetv2_wapv2_ce_sd2mfsf_uspg.pth"


class SFD2_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DETECT_DESCRIBE
    default_conf = {
        "top_k": 2048,
        "model_name": "ressegnetv2",  # "ressegnetv2", "ressegnet"
        "use_stability": True,
        "conf_th": 0.001,
        "scales": [1.0],
    }

    def __init__(self, conf={}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)

        self.max_kps = conf.top_k
        self.DEV = torch.device("cpu")
        self.model = SFD2(
            model_name=conf.model_name,
            use_stability=conf.use_stability,
            top_k=conf.top_k,
            conf_th=conf.conf_th,
            scales=conf.scales,
        )

        weights = downloadModel("sfd2", "20220810_ressegnetv2_wapv2_ce_sd2mfsf_uspg.pth", weights_link)
        self.model.model.load_state_dict(torch.load(weights, map_location="cpu")["model"], strict=False)
        self.matcher = NearestNeighborMatcher()

    def detectAndCompute(self, img, return_dict=False):
        img = ops.prepareImage(img)

        with torch.no_grad():
            pred = self.model(
                {
                    "image": img.to(self.DEV),
                    "original_size": torch.tensor(img.shape[-2:]).to(self.DEV),
                }
            )

        # make all into tensor and move to device
        for k in pred:
            pred[k] = torch.tensor(pred[k]).to(self.DEV).unsqueeze(0)

        if return_dict:
            return pred

        return pred["keypoints"], pred["descriptors"]

    def detect(self, img):
        raise NotImplementedError("This method is not implemented in this class")

    def compute(self, img, kps):
        raise NotImplementedError("This method is not implemented in this class")

    def to(self, device):
        self.model = self.model.to(device)
        self.DEV = device
        return self

    @property
    def has_detector(self):
        return True



import torch
import torch.nn.functional as F
from omegaconf import OmegaConf

from easy_local_features.submodules.git_roma import romatch

from .basemodel import BaseExtractor, MethodType
from typing import TypedDict


class ROMAConfig(TypedDict):
    model_name: str
    top_k: int
    detection_threshold: float
    nms_radius: int
    model: str
    upsample_factor: int


models = {
    "outdoor": romatch.models.roma_outdoor,
    "indoor": romatch.models.roma_indoor,
    "tiny_outdoor": romatch.models.tiny_roma_v1_outdoor,
}


class RoMa_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.END2END_MATCHER
    default_conf: ROMAConfig = {
        "model_name": "roma",
        "top_k": 512,
        "detection_threshold": 0.2,
        "nms_radius": 4,
        "model": "outdoor",
        "upsample_factor": 8,
    }

    def __init__(self, conf: ROMAConfig = {}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        self.num_keypoints = conf.top_k
        self.device = torch.device("cpu")
        self.model = models[self.conf.model](device=self.device)
        self.model.eval()

    def preprocess_batch_tensor(
        self,
        tensor: torch.Tensor,
        target_size: tuple[int, int],
        mean=(0.485, 0.456, 0.406),
        std=(0.229, 0.224, 0.225),
    ) -> torch.Tensor:
        if tensor.ndim == 3:
            tensor = tensor.unsqueeze(0)
        x = tensor
        hs, ws = target_size
        x = F.interpolate(x, size=(hs, ws), mode="bicubic", align_corners=False)
        dev = x.device
        m = torch.tensor(mean, device=dev).view(1, 3, 1, 1)
        s = torch.tensor(std, device=dev).view(1, 3, 1, 1)
        x = (x - m) / s
        return x

    def detectAndCompute(self, image, return_dict=False):
        raise NotImplementedError("Every BaseExtractor must implement the detectAndCompute method.")

    def detect(self, image):
        raise NotImplementedError("Every BaseExtractor must implement the detect method.")

    def compute(self, image, keypoints):
        raise NotImplementedError("Every BaseExtractor must implement the compute method.")

    def to(self, device):
        # RoMa model constructor accepts device; here we move logical state
        self.device = torch.device(device)
        # romatch API builds internal modules on init; no explicit .to available
        # but we can recreate the model on the target device to be safe
        model_key = self.conf.get("model", "outdoor")
        if model_key in models:
            self.model = models[model_key](device=self.device)
            self.model.eval()
        return self

    @property
    def has_detector(self):
        return True

    # @abstractmethod
    def match(self, image1, image2):
        hs, ws = self.model.h_resized, self.model.w_resized
        imA_h, imA_w = image1.shape[2:]
        imB_h, imB_w = image2.shape[2:]

        batchA = self.preprocess_batch_tensor(image1, (hs, ws)).to(self.device)
        batchB = self.preprocess_batch_tensor(image2, (hs, ws)).to(self.device)

        warp, certainty = self.model.match(batchA, batchB, batched=True, device=self.device)
        matches, certs = self.model.sample(warp, certainty, num=self.num_keypoints)

        kptsA, kptsB = self.model.to_pixel_coordinates(matches, imA_h, imA_w, imB_h, imB_w)
        # Ensure keypoints are detached and on CPU
        return {
            "mkpts0": kptsA.detach().cpu() if isinstance(kptsA, torch.Tensor) else kptsA,
            "mkpts1": kptsB.detach().cpu() if isinstance(kptsB, torch.Tensor) else kptsB,
        }



"""Pytest configuration to ensure the package in src/ is importable.

Without installing the project (editable or wheel), running `pytest` from the
repository root won't put the `src/` directory on `sys.path`, so
`import easy_local_features` fails with ModuleNotFoundError.

This file prepends the absolute path to `src/` so tests can import the package
when executed via `pytest` directly (e.g. `uv tool run pytest`).
"""

from __future__ import annotations
import sys
from pathlib import Path

# Path to this file: <repo>/src/test/conftest.py
# We want to add <repo>/src to sys.path
SRC_DIR = Path(__file__).resolve().parents[1]  # parents[1] == <repo>/src
if str(SRC_DIR) not in sys.path:
    sys.path.insert(0, str(SRC_DIR))



import cv2
import numpy as np
import torch
from omegaconf import OmegaConf

from ..matching.nearest_neighbor import NearestNeighborMatcher
from ..utils import ops
from .basemodel import BaseExtractor, MethodType
from typing import TypedDict


class ORBConfig(TypedDict):
    top_k: int
    scaleFactor: float
    nlevels: int
    edgeThreshold: int
    firstLevel: int
    WTA_K: int
    scoreType: int
    patchSize: int
    fastThreshold: int


class ORB_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DETECT_DESCRIBE
    default_conf: ORBConfig = {
        "top_k": 2048,
        "scaleFactor": 1.2,
        "nlevels": 8,
        "edgeThreshold": 31,
        "firstLevel": 0,
        "WTA_K": 2,
        "scoreType": cv2.ORB_HARRIS_SCORE,
        "patchSize": 31,
        "fastThreshold": 20,
    }

    def __init__(self, conf: ORBConfig = {}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        self.model = cv2.ORB_create(
            nfeatures=conf.top_k,
            scaleFactor=conf.scaleFactor,
            nlevels=conf.nlevels,
            edgeThreshold=conf.edgeThreshold,
            firstLevel=conf.firstLevel,
            WTA_K=conf.WTA_K,
            scoreType=conf.scoreType,
            patchSize=conf.patchSize,
            fastThreshold=conf.fastThreshold,
        )
        self.matcher = NearestNeighborMatcher()

    def detectAndCompute(self, img, return_dict=False):
        img = ops.prepareImage(img, gray=True, batch=False)
        img = ops.to_cv(img)

        keypoints, descriptors = self.model.detectAndCompute(img, None)
        # Convert to torch tensors with batch dimension
        kps_np = np.array([kp.pt for kp in keypoints], dtype=np.float32)
        desc_np = (
            np.array(descriptors, dtype=np.float32) if descriptors is not None else np.zeros((0, 32), dtype=np.float32)
        )
        kps = torch.from_numpy(kps_np).float().unsqueeze(0)
        desc = torch.from_numpy(desc_np).float().unsqueeze(0)
        pred = {
            "keypoints": kps,
            "descriptors": desc,
        }

        if return_dict:
            return pred
        return pred["keypoints"], pred["descriptors"]

    def detect(self, img):
        img = ops.prepareImage(img, gray=True, batch=False)
        img = ops.to_cv(img)

        keypoints = self.model.detect(img, None)
        kps_np = np.array([kp.pt for kp in keypoints], dtype=np.float32)
        kps = torch.from_numpy(kps_np).float().unsqueeze(0)
        return kps

    def compute(self, img, kps):
        raise NotImplementedError("This method is not implemented in this class")

    def to(self, device):
        return self

    @property
    def has_detector(self):
        return True



import torch


def kde(x, std = 0.1, half = True, down = None):
    # use a gaussian kernel to estimate density
    if half:
        x = x.half() # Do it in half precision TODO: remove hardcoding
    if down is not None:
        scores = (-torch.cdist(x,x[::down])**2/(2*std**2)).exp()
    else:
        scores = (-torch.cdist(x,x)**2/(2*std**2)).exp()
    density = scores.sum(dim=-1)
    return density


from __future__ import annotations

from dataclasses import dataclass
from typing import List, Sequence, Union

import numpy as np
import torch

from ..utils import ops
from .basemodel import ImageLike


DetectorLike = Union[object]


@dataclass
class EnsembleDetectorConfig:
    """Configuration for the `EnsembleDetector`.

    Args:
        deduplicate: If True, remove duplicate keypoints across detectors by
            rounding coordinates to the nearest pixel and uniquifying.
        sort: If True, sort final keypoints lexicographically for stable order.
        max_keypoints: Optional cap on total keypoints per image. When exceeded,
            a simple deterministic subsample (stride selection) is applied.
    """

    deduplicate: bool = True
    sort: bool = True
    max_keypoints: int | None = None


class EnsembleDetector:
    """Aggregate multiple detectors and expose a single `detect()` API.

    Typical usage:
        - Pass instantiated detectors (any object exposing `detect(image)` that
          returns `[1, N, 2]` or `[N, 2]` keypoints in pixel coordinates).
        - Call `detect(image)` with a single image or a batch (`[B,C,H,W]`, NumPy, or path).
        - Get a tensor `[B, N_total, 2]` (padded across batches) of (x, y) keypoints.

    Notes:
        - Each underlying detector is invoked per image (looping over batch dim)
          to support non-batched detectors (e.g., OpenCV ORB).
        - Deduplication rounds keypoints to integer pixel locations before uniquifying.
    """

    def __init__(self, detectors: Sequence[DetectorLike], config: EnsembleDetectorConfig | None = None):
        assert len(detectors) > 0, "EnsembleDetector requires at least one detector"
        self.detectors: List[DetectorLike] = list(detectors)
        self.config = config or EnsembleDetectorConfig()
        self.device = torch.device("cpu")

    def to(self, device: torch.device | str):
        """Move underlying detectors to a device when supported and return self."""
        for d in self.detectors:
            to_fn = getattr(d, "to", None)
            if callable(to_fn):
                try:
                    to_fn(device)
                except Exception:
                    # Some detectors (e.g., OpenCV) may not support .to(); ignore.
                    pass
        return self

    @torch.inference_mode()
    def detect(self, image: ImageLike) -> torch.Tensor:
        """Detect and merge keypoints from all detectors.

        Args:
            image: Path, numpy array, or torch tensor. Can be batched.

        Returns:
            torch.FloatTensor: `[B, N, 2]` keypoints (x, y) in pixel coords.
            For single images, `B==1`. Across a batch, per-image lists are padded
            to the maximum length in the batch.
        """
        # Normalize input to [B,C,H,W]
        img = ops.prepareImage(image, gray=False, batch=True)
        device = img.device if isinstance(img, torch.Tensor) else torch.device("cpu")

        B = img.shape[0]
        out_kps: List[torch.Tensor] = []

        for b in range(B):
            per_img_kps: List[torch.Tensor] = []
            single = img[b]
            for det in self.detectors:
                kps = det.detect(single)
                if not isinstance(kps, torch.Tensor):
                    kps = torch.as_tensor(kps)
                if kps.ndim == 3 and kps.shape[0] == 1:
                    kps = kps[0]
                assert kps.ndim == 2 and kps.shape[-1] == 2, (
                    f"Detector {type(det).__name__} returned keypoints with shape {tuple(kps.shape)}, expected [N,2] or [1,N,2]."
                )
                kps = kps.to(device=device, dtype=torch.float32)
                per_img_kps.append(kps)

            merged = per_img_kps[0] if len(per_img_kps) == 1 else torch.cat(per_img_kps, dim=0)

            if self.config.deduplicate and merged.numel() > 0:
                rounded = torch.round(merged).to(torch.int64)
                rounded_np = rounded.detach().cpu().numpy()
                _, first_idx = np.unique(rounded_np, axis=0, return_index=True)
                keep = torch.from_numpy(np.sort(first_idx)).to(merged.device)
                merged = merged[keep]

            if self.config.sort and merged.numel() > 0:
                merged_np = merged.detach().cpu().numpy()
                merged_sorted = ops.sort_keypoints(merged_np)
                merged = torch.from_numpy(merged_sorted).to(device=device, dtype=merged.dtype)

            if self.config.max_keypoints is not None and merged.shape[0] > self.config.max_keypoints:
                N = self.config.max_keypoints
                stride = max(1, merged.shape[0] // N)
                merged = merged[::stride][:N]

            out_kps.append(merged.unsqueeze(0))

        max_len = max(k.shape[1] for k in out_kps)
        if max_len == 0:
            return torch.zeros((B, 0, 2), dtype=torch.float32, device=device)

        padded = []
        for k in out_kps:
            if k.shape[1] < max_len:
                pad = torch.zeros((1, max_len - k.shape[1], 2), dtype=k.dtype, device=k.device)
                padded.append(torch.cat([k, pad], dim=1))
            else:
                padded.append(k)
        return torch.cat(padded, dim=0)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

# References:
#   https://github.com/facebookresearch/dino/blob/master/vision_transformer.py
#   https://github.com/rwightman/pytorch-image-models/tree/master/timm/models/vision_transformer.py

import logging

from torch import Tensor
from torch import nn


logger = logging.getLogger("dinov2")


try:
    from xformers.ops import memory_efficient_attention, unbind, fmha

    XFORMERS_AVAILABLE = True
except ImportError:
    # logger.warning("xFormers not available")
    XFORMERS_AVAILABLE = False


class Attention(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int = 8,
        qkv_bias: bool = False,
        proj_bias: bool = True,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim, bias=proj_bias)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: Tensor) -> Tensor:
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)

        q, k, v = qkv[0] * self.scale, qkv[1], qkv[2]
        attn = q @ k.transpose(-2, -1)

        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class MemEffAttention(Attention):
    def forward(self, x: Tensor, attn_bias=None) -> Tensor:
        if not XFORMERS_AVAILABLE:
            assert attn_bias is None, "xFormers is required for nested tensors usage"
            return super().forward(x)

        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads)

        q, k, v = unbind(qkv, 2)

        x = memory_efficient_attention(q, k, v, attn_bias=attn_bias)
        x = x.reshape([B, N, C])

        x = self.proj(x)
        x = self.proj_drop(x)
        return x



import importlib
import os
import pkgutil
from pathlib import Path

import pytest

from easy_local_features.feature.basemodel import BaseExtractor
from easy_local_features.utils import io, ops, vis

ROOT = Path(__file__).resolve().parent


def get_all_subclasses(cls):
    subclasses = set(cls.__subclasses__())
    for subclass in list(subclasses):
        subclasses.update(get_all_subclasses(subclass))
    return subclasses


def load_all_modules_from_package(package):
    for _, module_name, is_pkg in pkgutil.walk_packages(package.__path__, package.__name__ + "."):
        if not is_pkg:
            importlib.import_module(module_name)


@pytest.fixture
def all_subclasses():
    import easy_local_features

    load_all_modules_from_package(easy_local_features.feature)
    return get_all_subclasses(BaseExtractor)


@pytest.mark.skip
def test_feature_extractors(all_subclasses: list[BaseExtractor]):
    image0 = io.fromPath(str(ROOT / "assets/notredame.png"))
    image1 = io.fromPath(str(ROOT / "assets/notredame2.jpeg"))

    image0 = ops.crop_square(ops.resize_short_edge(image0, 320)[0])
    image1 = ops.crop_square(ops.resize_short_edge(image1, 320)[0])

    os.makedirs("test/results", exist_ok=True)

    for subclass in all_subclasses:
        if hasattr(subclass, "variations"):
            variations = subclass.variations
            for key, values in variations.items():
                for value in values:
                    try:
                        extractor = subclass({key: value, "top_k": 128})
                        matches = extractor.match(image0, image1)

                        vis.plot_pair(image0, image1, title=subclass.__name__ + " " + value, figsize=(8, 4))
                        vis.plot_matches(matches["mkpts0"], matches["mkpts1"])
                        vis.save(f"test/results/{subclass.__name__}_{value}.png")
                    except:
                        print(f"Error in {subclass.__name__} with {key}={value}")
        else:
            extractor = subclass({"top_k": 128})
            matches = extractor.match(image0, image1)

            vis.plot_pair(image0, image1, title=subclass.__name__, figsize=(8, 4))
            vis.plot_matches(matches["mkpts0"], matches["mkpts1"])
            vis.save(f"test/results/{subclass.__name__}.png")


if __name__ == "__main__":
    import argparse

    def parse():
        parser = argparse.ArgumentParser()
        parser.add_argument("--model", "-m", type=str, default="all", help="Model name to test. Default: all.")
        return parser.parse_args()

    args = parse()

    import easy_local_features.matching

    load_all_modules_from_package(easy_local_features.matching)
    _all_subclasses = get_all_subclasses(BaseExtractor)

    # print all subclasses
    for subclass in _all_subclasses:
        print(subclass.__name__)

    if args.model != "all":
        _all_subclasses = [subclass for subclass in _all_subclasses if args.model in subclass.__name__]

    test_feature_extractors(_all_subclasses)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Optional

from torch import Tensor, nn
import torch.nn.functional as F


class SwiGLUFFN(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: Optional[int] = None,
        out_features: Optional[int] = None,
        act_layer: Callable[..., nn.Module] = None,
        drop: float = 0.0,
        bias: bool = True,
    ) -> None:
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.w12 = nn.Linear(in_features, 2 * hidden_features, bias=bias)
        self.w3 = nn.Linear(hidden_features, out_features, bias=bias)

    def forward(self, x: Tensor) -> Tensor:
        x12 = self.w12(x)
        x1, x2 = x12.chunk(2, dim=-1)
        hidden = F.silu(x1) * x2
        return self.w3(hidden)


try:
    from xformers.ops import SwiGLU

    XFORMERS_AVAILABLE = True
except ImportError:
    SwiGLU = SwiGLUFFN
    XFORMERS_AVAILABLE = False


class SwiGLUFFNFused(SwiGLU):
    def __init__(
        self,
        in_features: int,
        hidden_features: Optional[int] = None,
        out_features: Optional[int] = None,
        act_layer: Callable[..., nn.Module] = None,
        drop: float = 0.0,
        bias: bool = True,
    ) -> None:
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        hidden_features = (int(hidden_features * 2 / 3) + 7) // 8 * 8
        super().__init__(
            in_features=in_features,
            hidden_features=hidden_features,
            out_features=out_features,
            bias=bias,
        )



import torch
import torchvision
import numpy as np
import cv2
from torch.nn import functional as F

from .io import fromPath

def prepareImage(image, gray=False, batch=True, imagenet=False):

    if isinstance(image, str):
        return fromPath(image, gray, batch, imagenet)
    
    if isinstance(image, np.ndarray):
        if image.dtype == np.uint8:
            image = image.astype(np.float32) / 255.0

        image = torch.from_numpy(image)

    if len(image.shape) == 2:
        image = image.unsqueeze(0)
    
    if len(image.shape) == 3 and image.shape[0] not in [1, 3]:
        image = image.permute(2, 0, 1)
        
    if len(image.shape) == 4 and image.shape[1] not in [1, 3]:
        image = image.permute(0, 3, 1, 2)
        
    if gray and image.shape[-3] == 3:
        image = image.mean(-3, keepdim=True)
        
    if imagenet:
        image = torchvision.transforms.Normalize(
            mean=[0.485, 0.456, 0.406],
            std=[0.229, 0.224, 0.225]
        )(image)
        
    if batch and len(image.shape) == 3:
        image = image.unsqueeze(0)
        
    if not batch and len(image.shape) == 4:
        image = image[0]
        
    return image


def to_cv(torch_image, convert_color=False, batch_idx=0, to_gray=False):
    '''Converts a torch tensor image to a numpy array'''
    
    if isinstance(torch_image, torch.Tensor):
        if len(torch_image.shape) == 2:
            torch_image = torch_image.unsqueeze(0)
        elif len(torch_image.shape) == 4:
            torch_image = torch_image[batch_idx]
            
        if torch_image.max() > 1:
            torch_image = torch_image / torch_image.max()
        
        img = (torch_image.permute(1, 2, 0).detach().cpu().numpy() * 255).astype("uint8")
    else:
        img = torch_image

    if convert_color:
        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
        
    if to_gray:
        if len(img.shape) == 3:
            if img.shape[-1] == 3:
                img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        elif len(img.shape) == 2:
            img = img[None]
        
    return img

def resize_short_edge(image, min_size):
    h, w = image.shape[-2:]
    if h > w:
        new_w = min_size
        new_h = int(h * min_size / w)
    else:
        new_h = min_size
        new_w = int(w * min_size / h)
    
    scale = new_h / h

    if image.dim() == 2:
        return torch.nn.functional.interpolate(image.unsqueeze(0).unsqueeze(0), size=(new_h, new_w), mode='nearest').squeeze(0).squeeze(0), scale
    
    if image.dim() == 3:
        return torch.nn.functional.interpolate(image.unsqueeze(0), size=(new_h, new_w), mode='bilinear', align_corners=False).squeeze(0), scale
    
    return torch.nn.functional.interpolate(image, size=(new_h, new_w), mode='bilinear', align_corners=False), scale

def pad_square(image):
    # pad bottom or right
    h, w = image.shape[-2:]
    if h > w:
        pad = h - w
        image = torch.nn.functional.pad(image, (0, pad, 0, 0), mode='constant', value=0)
    else:
        pad = w - h
        image = torch.nn.functional.pad(image, (0, 0, 0, pad), mode='constant', value=0)
    return image

def crop_square(image):
    # crop bottom or right
    h, w = image.shape[-2:]
    if h == w:
        return image
    if h > w:
        crop = h - w
        image = image[...,:-crop, :]
    else:
        crop = w - h
        image = image[..., :-crop]
    return image

def crop_patches(image, keypoints, patch_size = 32, mode='nearest'):
    B, C, H, W = image.shape
    N = keypoints.shape[1]

    # Ensure the keypoints are in the correct range
    x_coords = keypoints[:, :, 0]
    y_coords = keypoints[:, :, 1]
    
    # Calculate the left and top edges of the patches
    left = x_coords - patch_size / 2
    top = y_coords - patch_size / 2

    # Normalize to [-1, 1] for grid_sample
    left_norm = 2 * left / (W - 1) - 1
    top_norm = 2 * top / (H - 1) - 1

    # Create the sampling grid
    # grid_x = torch.linspace(-1, 1, patch_size).to(image.device)
    # grid_y = torch.linspace(-1, 1, patch_size).to(image.device)
    # this is not from -1 to 1, bt from the left to the right in normalized coordinates
    patch_size_normalized_x = patch_size / W
    patch_size_normalized_y = patch_size / H
    grid_x = torch.linspace(-patch_size_normalized_x, patch_size_normalized_x, patch_size).to(image.device)
    grid_y = torch.linspace(-patch_size_normalized_y, patch_size_normalized_y, patch_size).to(image.device)
    
    grid_x, grid_y = torch.meshgrid(grid_x, grid_y, indexing='ij')
    grid = torch.stack((grid_x, grid_y), 2).unsqueeze(0)  # 1xpatch_sizexpatch_sizex2

    grid = grid.repeat(B, N, 1, 1, 1)  # BxNxpatch_sizexpatch_sizex2

    # Adjust the grid according to keypoints
    grid[:, :, :, :, 0] = (grid[:, :, :, :, 0] + left_norm.unsqueeze(2).unsqueeze(3))
    grid[:, :, :, :, 1] = (grid[:, :, :, :, 1] + top_norm.unsqueeze(2).unsqueeze(3))
    
    # the patch is anchored at the bottom right corner of the keypoint
    # let's move it to the center of the patch
    grid[:, :, :, :, 0] += patch_size_normalized_x
    grid[:, :, :, :, 1] += patch_size_normalized_y
    
    # Reshape for grid_sample
    grid = grid.view(B * N, patch_size, patch_size, 2)

    # Repeat images and apply grid_sample
    image_repeated = image.unsqueeze(1).repeat(1, N, 1, 1, 1).view(B * N, C, H, W)
    patches = F.grid_sample(image_repeated, grid, mode=mode, align_corners=True)

    patches = patches.view(B, N, C, patch_size, patch_size)
    # flip the patches to have the same orientation as the other methods
    # transpose the last two dimensions
    patches = patches.transpose(-1, -2)
    
    return patches


def sort_keypoints(mkpts0, mkpts1=None):
    # get a lexical index of the keypoints based on the x and y coordinates
    if len(mkpts0.shape) == 3:
        # batched keypoints
        B, N, _ = mkpts0.shape
        for b in range(B):
            idxs = np.lexsort((mkpts0[b, :, 0], mkpts0[b, :, 1]))
            mkpts0[b] = mkpts0[b, idxs]
            if mkpts1 is not None:
                mkpts1[b] = mkpts1[b, idxs]

        if mkpts1 is not None:
            return mkpts0, mkpts1
        return mkpts0
    else:
        idxs = np.lexsort((mkpts0[:, 0], mkpts0[:, 1]))
        mkpts0 = mkpts0[idxs]
        if mkpts1 is not None:
            mkpts1 = mkpts1[idxs]
            return mkpts0, mkpts1

        return mkpts0    
    
def to_homogeneous(kpts):
    if isinstance(kpts, np.ndarray):
        if len(kpts.shape) == 3:
            B, N, _ = kpts.shape
            return np.concatenate((kpts, np.ones((B, N, 1))), axis=-1)
        else:
            return np.concatenate((kpts, np.ones((kpts.shape[0], 1))), axis=-1)
    elif isinstance(kpts, torch.Tensor):
        if len(kpts.shape) == 3:
            B, N, _ = kpts.shape
            return torch.cat((kpts, torch.ones(B, N, 1, device=kpts.device)), dim=-1)
        else:
            return torch.cat((kpts, torch.ones(kpts.shape[0], 1, device=kpts.device)), dim=-1)
        
def to_grid(kpts, H, W):
    # kpts => in pixel coordinates
    # H, W => image dimensions
    # returns the grid coordinates (-1, 1) for grid_sample
    
    if isinstance(kpts, np.ndarray):
        if len(kpts.shape) == 3:
            B, N, _ = kpts.shape
            return 2 * kpts / np.array([W, H]) - 1
        else:
            return 2 * kpts / np.array([W, H]) - 1
        
    elif isinstance(kpts, torch.Tensor):
        if len(kpts.shape) == 3:
            B, N, _ = kpts.shape
            return 2 * kpts / torch.tensor([W, H], device=kpts.device) - 1
        else:
            return 2 * kpts / torch.tensor([W, H], device=kpts.device) - 1
        
def from_homogeneous(kpts):
    if isinstance(kpts, np.ndarray):
        if len(kpts.shape) == 3:
            B, N, _ = kpts.shape
            return kpts[:, :, :2] / kpts[:, :, 2:]
        else:
            return kpts[:, :2] / kpts[:, 2:]
    elif isinstance(kpts, torch.Tensor):
        if len(kpts.shape) == 3:
            B, N, _ = kpts.shape
            return kpts[:, :, :2] / kpts[:, :, 2:]
        else:
            return kpts[:, :2] / kpts[:, 2:]
        
@torch.no_grad()
def warp_kpts(kpts0, depth0, depth1, T_0to1, K0, K1, do_on_cpu=False):
    """ Warp kpts0 from I0 to I1 with depth, K and Rt
    Also check covisibility and depth consistency.
    Depth is consistent if relative error < 0.2 (hard-coded).
    
    Args:
        kpts0 (torch.Tensor): [N, L, 2] - <x, y>,
        depth0 (torch.Tensor): [N, H, W],
        depth1 (torch.Tensor): [N, H, W],
        T_0to1 (torch.Tensor): [N, 3, 4],
        K0 (torch.Tensor): [N, 3, 3],
        K1 (torch.Tensor): [N, 3, 3],
    Returns:
        calculable_mask (torch.Tensor): [N, L]
        warped_keypoints0 (torch.Tensor): [N, L, 2] <x0_hat, y1_hat>
    """
    if do_on_cpu:
        og_device = kpts0.device
        kpts0 = kpts0.cpu()
        depth0 = depth0.cpu()
        depth1 = depth1.cpu()
        T_0to1 = T_0to1.cpu()
        K0 = K0.cpu()
        K1 = K1.cpu()
    
    kpts0_long = (kpts0 - 0.5).round().long().clip(0, 2000-1)

    depth0[:, 0, :] = 0 ; depth1[:, 0, :] = 0 
    depth0[:, :, 0] = 0 ; depth1[:, :, 0] = 0 

    # Sample depth, get calculable_mask on depth != 0
    kpts0_depth = torch.stack(
        [depth0[i, kpts0_long[i, :, 1], kpts0_long[i, :, 0]] for i in range(kpts0.shape[0])], dim=0
    )  # (N, L)
    nonzero_mask = kpts0_depth > 0

    # Unproject
    kpts0_h = torch.cat([kpts0, torch.ones_like(kpts0[:, :, [0]])], dim=-1) * kpts0_depth[..., None]  # (N, L, 3)
    kpts0_cam = K0.inverse() @ kpts0_h.transpose(2, 1)  # (N, 3, L)

    # Rigid Transform
    w_kpts0_cam = T_0to1[:, :3, :3] @ kpts0_cam + T_0to1[:, :3, [3]]    # (N, 3, L)
    w_kpts0_depth_computed = w_kpts0_cam[:, 2, :] 

    # Project
    w_kpts0_h = (K1 @ w_kpts0_cam).transpose(2, 1)  # (N, L, 3)
    w_kpts0 = w_kpts0_h[:, :, :2] / (w_kpts0_h[:, :, [2]] + 1e-5)  # (N, L, 2), +1e-4 to avoid zero depth
    
    kpts1_depth = torch.zeros_like(kpts0_depth)
    
    w_kpts0_long = w_kpts0.round().long()
    for i in range(w_kpts0.shape[0]):
        valid = (w_kpts0_long[i, :, 0] >= 0) & (w_kpts0_long[i, :, 0] < depth1.shape[2]) & (w_kpts0_long[i, :, 1] >= 0) & (w_kpts0_long[i, :, 1] < depth1.shape[1])
        kpts1_depth[i, valid] = depth1[i, w_kpts0_long[i, valid, 1], w_kpts0_long[i, valid, 0]]

    depth_diff = torch.abs(w_kpts0_depth_computed - kpts1_depth)
    depth_diff_mask = depth_diff < 0.2
    
    inside_the_image_mask = (w_kpts0[..., 0] > 0) & (w_kpts0[..., 0] < depth1.shape[2]) & (w_kpts0[..., 1] > 0) & (w_kpts0[..., 1] < depth1.shape[1])
    
    valid_mask = nonzero_mask * depth_diff_mask * inside_the_image_mask
    
    if do_on_cpu:
        valid_mask = valid_mask.to(og_device)
        w_kpts0 = w_kpts0.to(og_device)
        kps0 = kpts0.to(og_device)
        depth0 = depth0.to(og_device)
        depth1 = depth1.to(og_device)
        T_0to1 = T_0to1.to(og_device)
        K0 = K0.to(og_device)
        K1 = K1.to(og_device)

    return valid_mask, w_kpts0


from setuptools import setup, find_packages

setup(
    name="romatch",
    packages=find_packages(include=("romatch*",)),
    version="0.0.2",
    author="Johan Edstedt",
    install_requires=open("requirements.txt", "r").read().split("\n"),
)



import os.path
from typing import Tuple

import torch.nn.functional as F

from .utils import *

__all__ = [
    'e2wrn10_8R',    
    'e2wrn16_8R_stl',
    'e2wrn16_8_stl',
    'e2wrn28_10',
    'e2wrn28_7',
    'e2wrn28_10R',
    'e2wrn28_7R',
    'e2wrn28_10_gpool',
]

model_paths = {
    'e2wrn10_8R': os.path.join(STORE_PATH, 'e2wrn10-8R.model'),    
    'e2wrn16_8R_stl': os.path.join(STORE_PATH, 'e2wrn16-8R_stl.model'),
    'e2wrn16_8_stl': os.path.join(STORE_PATH, 'e2wrn16-8_stl.model'),
    'e2wrn28_10': os.path.join(STORE_PATH, 'e2wrn28-10.model'),
    'e2wrn28_7': os.path.join(STORE_PATH, 'e2wrn28-7.model'),
    'e2wrn28_10R': os.path.join(STORE_PATH, 'e2wrn28-10R.model'),
    'e2wrn28_7R': os.path.join(STORE_PATH, 'e2wrn28-7R.model'),
    'e2wrn28_10_gpool': os.path.join(STORE_PATH, 'e2wrn28_10_gpool.model'),
}

########################################################################################################################
# Code adapted from:
# https://github.com/meliketoy/wide-resnet.pytorch/blob/master/networks/wide_resnet.py
# which has the following MIT License:
########################################################################################################################
# MIT License
#
# Copyright (c) 2018 Bumsoo Kim
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
########################################################################################################################
import torch


class WideBasic(nn.EquivariantModule):
    
    def __init__(self,
                 in_fiber: nn.FieldType,
                 inner_fiber: nn.FieldType,
                 dropout_rate, stride=1,
                 out_fiber: nn.FieldType = None,
                 F: float = 1.,
                 sigma: float = 0.45,
                 ):
        super(WideBasic, self).__init__()
        
        if out_fiber is None:
            out_fiber = in_fiber
        
        self.in_type = in_fiber
        inner_class = inner_fiber
        self.out_type = out_fiber
        
        if isinstance(in_fiber.gspace, gspaces.FlipRot2dOnR2):
            rotations = in_fiber.gspace.fibergroup.rotation_order
        elif isinstance(in_fiber.gspace, gspaces.Rot2dOnR2):
            rotations = in_fiber.gspace.fibergroup.order()
        else:
            rotations = 0
        
        if rotations in [0, 2, 4]:
            conv = conv3x3
        else:
            conv = conv5x5
        
        self.bn1 = nn.InnerBatchNorm(self.in_type)
        self.relu1 = nn.ReLU(self.in_type, inplace=True)
        self.conv1 = conv(self.in_type, inner_class, sigma=sigma, F=F, initialize=False)
        
        self.bn2 = nn.InnerBatchNorm(inner_class)
        self.relu2 = nn.ReLU(inner_class, inplace=True)
        
        self.dropout = nn.PointwiseDropout(inner_class, p=dropout_rate)
        
        self.conv2 = conv(inner_class, self.out_type, stride=stride, sigma=sigma, F=F, initialize=False)
        
        self.shortcut = None
        if stride != 1 or self.in_type != self.out_type:
            self.shortcut = conv1x1(self.in_type, self.out_type, stride=stride, bias=False, sigma=sigma, F=F, initialize=False)
            # if rotations in [0, 2, 4]:
            #     self.shortcut = conv1x1(self.in_type, self.out_type, stride=stride, bias=False, sigma=sigma, F=F, initialize=False)
            # else:
            #     self.shortcut = conv3x3(self.in_type, self.out_type, stride=stride, bias=False, sigma=sigma, F=F, initialize=False)
    
    def forward(self, x):
        x_n = self.relu1(self.bn1(x))
        out = self.relu2(self.bn2(self.conv1(x_n)))
        out = self.dropout(out)
        out = self.conv2(out)
        
        if self.shortcut is not None:
            out += self.shortcut(x_n)
        else:
            out += x
        
        return out
    
    def evaluate_output_shape(self, input_shape: Tuple):
        assert len(input_shape) == 4
        assert input_shape[1] == self.in_type.size
        if self.shortcut is not None:
            return self.shortcut.evaluate_output_shape(input_shape)
        else:
            return input_shape


class Wide_ResNet(torch.nn.Module):
    def __init__(self, depth, widen_factor, dropout_rate, num_classes=100,
                 N: int = 8,
                 r: int = 1,
                 f: bool = True,
                 main_fiber: str = "regular",
                 inner_fiber: str = "regular",
                 F: float = 1.,
                 sigma: float = 0.45,
                 deltaorth: bool = False,
                 fixparams: bool = True,
                 initial_stride: int = 1,
                 conv2triv: bool = True,
                 ):
        super(Wide_ResNet, self).__init__()
        
        assert ((depth - 4) % 6 == 0), 'Wide-resnet depth should be 6n+4'
        n = int((depth - 4) / 6)
        k = widen_factor
        
        print(f'| Wide-Resnet {depth}x{k} ({CHANNELS_CONSTANT * 100}%)')
        
        nStages = [16, 16 * k, 32 * k, 64 * k]
        
        self.distributed = False
        self._fixparams = fixparams
        self.conv2triv = conv2triv
        
        self._layer = 0
        self._N = N
        
        # if the model is [F]lip equivariant
        self._f = f
        
        # level of [R]estriction:
        #   r < 0 : never do restriction, i.e. initial group (either D8 or C8) preserved for the whole network
        #   r = 0 : do restriction before first layer, i.e. initial group doesn't have rotation equivariance (C1 or D1)
        #   r > 0 : restrict after every block, i.e. start with 8 rotations, then restrict to 4 and finally 1
        self._r = r
        
        self._F = F
        self._sigma = sigma
        
        if self._f:
            self.gspace = gspaces.FlipRot2dOnR2(N)
            print("Successfully init reflection+rotation group.\n")

            print(" [Warning] Do not use reflection+rotation.")
            # raise NotImplementedError
        else:
            self.gspace = gspaces.Rot2dOnR2(N)
            print("Successfully init rotation group.\n") 


        if self._r == 0:
            id = (0, 1) if self._f else 1
            self.gspace, _, _ = self.gspace.restrict(id)
        
        r1 = nn.FieldType(self.gspace, [self.gspace.trivial_repr] * 3)
        self.in_type = r1
        
        # r2 = FIBERS[main_fiber](self.gspace, nStages[0], fixparams=self._fixparams)
        r2 = FIBERS[main_fiber](self.gspace, nStages[0], fixparams=True)
        self._in_type = r2
        
        self.conv1 = conv5x5(r1, r2, sigma=sigma, F=F, initialize=False)
        self.layer1 = self._wide_layer(WideBasic, nStages[1], n, dropout_rate, stride=initial_stride,
                                       main_fiber=main_fiber,
                                       inner_fiber=inner_fiber)
        if self._r > 0:
            id = (0, 4) if self._f else 4
            self.restrict1 = self._restrict_layer(id)
        else:
            self.restrict1 = lambda x: x
        
        self.layer2 = self._wide_layer(WideBasic, nStages[2], n, dropout_rate, stride=2,
                                       main_fiber=main_fiber,
                                       inner_fiber=inner_fiber)
        if self._r > 1:
            id = (0, 1) if self._f else 1
            self.restrict2 = self._restrict_layer(id)
        else:
            self.restrict2 = lambda x: x
        
        if self.conv2triv:
            out_fiber = "trivial"
        else:
            out_fiber = None
            
        self.layer3 = self._wide_layer(WideBasic, nStages[3], n, dropout_rate, stride=2,
                                       main_fiber=main_fiber,
                                       inner_fiber=inner_fiber,
                                       out_fiber=out_fiber
                                       )
        
        self.bn1 = nn.InnerBatchNorm(self.layer3.out_type, momentum=0.9)
        if self.conv2triv:
            self.relu = nn.ReLU(self.bn1.out_type, inplace=True)
        else:
            self.mp = nn.GroupPooling(self.layer3.out_type)
            self.relu = nn.ReLU(self.mp.out_type, inplace=True)
            
        self.linear = torch.nn.Linear(self.relu.out_type.size, num_classes)
        
        for name, module in self.named_modules():
            if isinstance(module, nn.R2Conv):
                if deltaorth:
                    init.deltaorthonormal_init(module.weights.data, module.basisexpansion)
                else:
                    init.generalized_he_init(module.weights.data, module.basisexpansion)
            elif isinstance(module, torch.nn.BatchNorm2d):
                module.weight.data.fill_(1)
                module.bias.data.zero_()
            elif isinstance(module, torch.nn.Linear):
                module.bias.data.zero_()
        
        # print("MODEL TOPOLOGY:")
        # for i, (name, mod) in enumerate(self.named_modules()):
        #     print(f"\t{i} - {name}")
        # for i, (name, mod) in enumerate(self.named_modules()):
        #     params = sum([p.numel() for p in mod.parameters() if p.requires_grad])
        #     if isinstance(mod, nn.EquivariantModule) and isinstance(mod.in_type, nn.FieldType) and isinstance(mod.out_type,
        #                                                                                                 nn.FieldType):
        #         print(f"\t{i: <3} - {name: <70} | {params: <8} | {mod.in_type.size: <4}- {mod.out_type.size: <4}")
        #     else:
        #         print(f"\t{i: <3} - {name: <70} | {params: <8} |")
        tot_param = sum([p.numel() for p in self.parameters() if p.requires_grad])
        print("Total number of parameters:", tot_param)


    def _restrict_layer(self, subgroup_id):
        layers = list()
        layers.append(nn.RestrictionModule(self._in_type, subgroup_id))
        layers.append(nn.DisentangleModule(layers[-1].out_type))
        self._in_type = layers[-1].out_type
        self.gspace = self._in_type.gspace
        
        restrict_layer = nn.SequentialModule(*layers)
        return restrict_layer
    
    def _wide_layer(self, block, planes: int, num_blocks: int, dropout_rate: float, stride: int,
                    main_fiber: str = "regular",
                    inner_fiber: str = "regular",
                    out_fiber: str = None,
                    ):
        
        self._layer += 1
        print("start building", self._layer)
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        
        main_type = FIBERS[main_fiber](self.gspace, planes, fixparams=self._fixparams)
        inner_class = FIBERS[inner_fiber](self.gspace, planes, fixparams=self._fixparams)
        if out_fiber is None:
            out_fiber = main_fiber
        out_type = FIBERS[out_fiber](self.gspace, planes, fixparams=self._fixparams)
        
        for b, stride in enumerate(strides):
            if b == num_blocks - 1:
                out_f = out_type
            else:
                out_f = main_type
            layers.append(
                block(self._in_type, inner_class, dropout_rate, stride, out_fiber=out_f, sigma=self._sigma, F=self._F))
            self._in_type = out_f
        print("built", self._layer)
        return nn.SequentialModule(*layers)
    
    def features(self, x):
        
        x = nn.GeometricTensor(x, self.in_type)
        
        out = self.conv1(x)
        
        x1 = self.layer1(out)
        
        if self.distributed:
            x1.tensor = x1.tensor.cuda(1)
        
        x2 = self.layer2(self.restrict1(x1))
        
        if self.distributed:
            x2.tensor = x2.tensor.cuda(2)
        
        x3 = self.layer3(self.restrict2(x2))
        # out = self.relu(self.mp(self.bn1(out)))
        
        return x1, x2, x3
    
    def forward(self, x):
        
        x = nn.GeometricTensor(x, self.in_type)
        
        out = self.conv1(x)
        out = self.layer1(out)
        
        if self.distributed:
            out.tensor = out.tensor.cuda(1)
        
        out = self.layer2(self.restrict1(out))
        
        if self.distributed:
            out.tensor = out.tensor.cuda(2)
        
        out = self.layer3(self.restrict2(out))
        
        if self.distributed:
            out.tensor = out.tensor.cuda(3)
        
        out = self.bn1(out)
        if not self.conv2triv:
            out = self.mp(out)
        out = self.relu(out)
        
        out = out.tensor
        
        b, c, w, h = out.shape
        out = F.avg_pool2d(out, (w, h))
        
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        
        return out
    
    def distribute(self):
        
        self.distributed = True
        
        self.conv1 = self.conv1.cuda(0)
        self.layer1 = self.layer1.cuda(0)
        
        if self._r:
            self.restrict1 = self.restrict1.cuda(1)
        self.layer2 = self.layer2.cuda(1)
        
        if self._r:
            self.restrict2 = self.restrict2.cuda(2)
        self.layer3 = self.layer3.cuda(2)
        
        self.relu = self.relu.cuda(3)
        self.bn1 = self.bn1.cuda(3)
        # self.mp = self.mp.cuda(3)
        self.avgpool = self.avgpool.cuda(3)
        self.linear = self.linear.cuda(3)
        
        return self


def e2wrn10_8R(pretrained=False, **kwargs):
    """Constructs a Wide ResNet 16-8 model with initial stride of 2 as mentioned here:
    https://github.com/uoguelph-mlrg/Cutout/issues/2

    Args:
        pretrained (bool): If True, returns a model pre-trained on STL10
    """
    model = Wide_ResNet(10, 8, 0, f=False, initial_stride=2, **kwargs)  ## No dropout
    if pretrained:
        model.load_state_dict(model_paths['e2wrn10_8R'])
    return model

def e2wrn16_8R_stl(pretrained=False, **kwargs):
    """Constructs a Wide ResNet 16-8 model with initial stride of 2 as mentioned here:
    https://github.com/uoguelph-mlrg/Cutout/issues/2

    Args:
        pretrained (bool): If True, returns a model pre-trained on STL10
    """
    model = Wide_ResNet(16, 8, 0, f=False, initial_stride=2, **kwargs)
    if pretrained:
        model.load_state_dict(model_paths['e2wrn16_8R_stl'])
    return model

def e2wrn16_8_stl(pretrained=False, **kwargs):
    """Constructs a Wide ResNet 16-8 model with initial stride of 2 as mentioned here:
    https://github.com/uoguelph-mlrg/Cutout/issues/2

    Args:
        pretrained (bool): If True, returns a model pre-trained on STL10
    """
    model = Wide_ResNet(16, 8, 0.3, f=True, initial_stride=2, **kwargs)
    if pretrained:
        model.load_state_dict(model_paths['e2wrn16_8_stl'])
    return model


def e2wrn28_10(pretrained=False, **kwargs):
    """Constructs a Wide ResNet 28-10 model

    Args:
        pretrained (bool): If True, returns a model pre-trained
    """
    model = Wide_ResNet(28, 10, 0.3, f=True, initial_stride=1, **kwargs)
    if pretrained:
        model.load_state_dict(model_paths['e2wrn28_10'])
    return model


def e2wrn28_7(pretrained=False, **kwargs):
    """Constructs a Wide ResNet 28-7 model

    Args:
        pretrained (bool): If True, returns a model pre-trained
    """
    model = Wide_ResNet(28, 7, 0.3, f=True, initial_stride=1, **kwargs)
    if pretrained:
        model.load_state_dict(model_paths['e2wrn28_7'])
    return model



def e2wrn28_10R(pretrained=False, **kwargs):
    """Constructs a Wide ResNet 28-10 model.
    This model is only [R]otation equivariant (no flips equivariance)

    Args:
        pretrained (bool): If True, returns a model pre-trained on Cifar100
    """
    model = Wide_ResNet(28, 10, 0.3, f=False, initial_stride=1, **kwargs)
    if pretrained:
        model.load_state_dict(model_paths['e2wrn28_10R'])
    return model


def e2wrn28_7R(pretrained=False, **kwargs):
    """Constructs a Wide ResNet 28-10 model.
    This model is only [R]otation equivariant (no flips equivariance)

    Args:
        pretrained (bool): If True, returns a model pre-trained on Cifar100
    """
    model = Wide_ResNet(28, 7, 0.3, f=False, initial_stride=1, **kwargs)
    if pretrained:
        model.load_state_dict(model_paths['e2wrn28_7R'])
    return model


def e2wrn28_10_gpool(pretrained=False, **kwargs):
    """Constructs a Wide ResNet 28-10 model with group pooling in the end

    Args:
        pretrained (bool): If True, returns a model pre-trained
    """
    model = Wide_ResNet(28, 10, 0.3, f=True, initial_stride=1, conv2triv=False, **kwargs)
    if pretrained:
        model.load_state_dict(model_paths['e2wrn28_10'])
    return model




import os, struct
import numpy as np
import zlib
import imageio
import cv2
import png
from tqdm import tqdm

from omegaconf import OmegaConf
import torch
import argparse
import h5py
from reasoning.modules.utils import load_image, load_depth
from reasoning.datasets.utils import *
from reasoning.features.desc_reasoning import sample_features_dino
from glob import glob

COMPRESSION_TYPE_COLOR = {-1:'unknown', 0:'raw', 1:'png', 2:'jpeg'}
COMPRESSION_TYPE_DEPTH = {-1:'unknown', 0:'raw_ushort', 1:'zlib_ushort', 2:'occi_ushort'}

import logging
class RGBDFrame:
    def load(self, file_handle):
        self.camera_to_world = np.asarray(struct.unpack('f'*16, file_handle.read(16*4)), dtype=np.float32).reshape(4, 4)
        self.timestamp_color = struct.unpack('Q', file_handle.read(8))[0]
        self.timestamp_depth = struct.unpack('Q', file_handle.read(8))[0]
        self.color_size_bytes = struct.unpack('Q', file_handle.read(8))[0]
        self.depth_size_bytes = struct.unpack('Q', file_handle.read(8))[0]
        self.color_data = b''.join(struct.unpack('c'*self.color_size_bytes, file_handle.read(self.color_size_bytes)))
        self.depth_data = b''.join(struct.unpack('c'*self.depth_size_bytes, file_handle.read(self.depth_size_bytes)))


    def decompress_depth(self, compression_type):
        if compression_type == 'zlib_ushort':
            return self.decompress_depth_zlib()
        else:
            raise


    def decompress_depth_zlib(self):
          return zlib.decompress(self.depth_data)


    def decompress_color(self, compression_type):
        if compression_type == 'jpeg':
            return self.decompress_color_jpeg()
        else:
            raise


    def decompress_color_jpeg(self):
        return imageio.imread(self.color_data)

class SensorData:
    def __init__(self, filename):
        self.version = 4
        self.load(filename)

    def load(self, filename):
        with open(filename, 'rb') as f:
            version = struct.unpack('I', f.read(4))[0]
            assert self.version == version
            strlen = struct.unpack('Q', f.read(8))[0]
            tmp = struct.unpack('c'*strlen, f.read(strlen))
            self.sensor_name = ''.join(map(lambda x: x.decode('utf-8'), tmp))
            self.intrinsic_color = np.asarray(struct.unpack('f'*16, f.read(16*4)), dtype=np.float32).reshape(4, 4)
            self.extrinsic_color = np.asarray(struct.unpack('f'*16, f.read(16*4)), dtype=np.float32).reshape(4, 4)
            self.intrinsic_depth = np.asarray(struct.unpack('f'*16, f.read(16*4)), dtype=np.float32).reshape(4, 4)
            self.extrinsic_depth = np.asarray(struct.unpack('f'*16, f.read(16*4)), dtype=np.float32).reshape(4, 4)
            self.color_compression_type = COMPRESSION_TYPE_COLOR[struct.unpack('i', f.read(4))[0]]
            self.depth_compression_type = COMPRESSION_TYPE_DEPTH[struct.unpack('i', f.read(4))[0]]
            self.color_width = struct.unpack('I', f.read(4))[0]
            self.color_height =  struct.unpack('I', f.read(4))[0]
            self.depth_width = struct.unpack('I', f.read(4))[0]
            self.depth_height =  struct.unpack('I', f.read(4))[0]
            self.depth_shift =  struct.unpack('f', f.read(4))[0]
            num_frames =  struct.unpack('Q', f.read(8))[0]
            self.frames = []
            for i in tqdm(range(num_frames), desc='Loading frames'):
                frame = RGBDFrame()
                frame.load(f)
                self.frames.append(frame)

    def export_depth_images(self, output_path, image_size=None, frame_skip=1):
        if not os.path.exists(output_path):os.makedirs(output_path)
        print('exporting', len(self.frames)//frame_skip, ' depth frames to', output_path)
        for f in tqdm(range(0, len(self.frames), frame_skip), desc='Exporting depth images'):
            depth_data = self.frames[f].decompress_depth(self.depth_compression_type)
            depth = np.fromstring(depth_data, dtype=np.uint16).reshape(self.depth_height, self.depth_width)
            if image_size is not None:depth = cv2.resize(depth, (image_size[1], image_size[0]), interpolation=cv2.INTER_NEAREST)
        #imageio.imwrite(os.path.join(output_path, str(f) + '.png'), depth)
            with open(os.path.join(output_path, str(f) + '.png'), 'wb') as f: # write 16-bit
                writer = png.Writer(width=depth.shape[1], height=depth.shape[0], bitdepth=16)
                depth = depth.reshape(-1, depth.shape[1]).tolist()
                writer.write(f, depth)

    def export_color_images(self, output_path, image_size=None, frame_skip=1):
        if not os.path.exists(output_path):os.makedirs(output_path)
        print('exporting', len(self.frames)//frame_skip, 'color frames to', output_path)
        for f in tqdm(range(0, len(self.frames), frame_skip), desc='Exporting color images'):
            color = self.frames[f].decompress_color(self.color_compression_type)
            if image_size is not None:
                color = cv2.resize(color, (image_size[1], image_size[0]), interpolation=cv2.INTER_NEAREST)
            imageio.imwrite(os.path.join(output_path, str(f) + '.jpg'), color)


    def save_mat_to_file(self, matrix, filename):
        with open(filename, 'w') as f:
            for line in matrix:
                np.savetxt(f, line[np.newaxis], fmt='%f')


    def export_poses(self, output_path, frame_skip=1):
        if not os.path.exists(output_path):os.makedirs(output_path)
        print('exporting', len(self.frames)//frame_skip, 'camera poses to', output_path)
        for f in tqdm(range(0, len(self.frames), frame_skip), desc='Exporting poses'):
            self.save_mat_to_file(self.frames[f].camera_to_world, os.path.join(output_path, str(f) + '.txt'))


    def export_intrinsics(self, output_path):
        if not os.path.exists(output_path):os.makedirs(output_path)
        print('exporting camera intrinsics to', output_path)
        self.save_mat_to_file(self.intrinsic_color, os.path.join(output_path, 'intrinsic_color.txt'))
        self.save_mat_to_file(self.extrinsic_color, os.path.join(output_path, 'extrinsic_color.txt'))
        self.save_mat_to_file(self.intrinsic_depth, os.path.join(output_path, 'intrinsic_depth.txt'))
        self.save_mat_to_file(self.extrinsic_depth, os.path.join(output_path, 'extrinsic_depth.txt'))

class ScannetImages(torch.utils.data.Dataset):
    defalut_config = {
        "data_path": "./datasets/scannet/scans/scene0000_00",
    }
    
    def __init__(self, config={}):
        self.config = config = OmegaConf.merge(OmegaConf.create(self.defalut_config), config)
        self.root = config.data_path
        # lexical sort
        self.all_color_files = sorted(os.listdir(os.path.join(self.root, 'color')), key=lambda x: int(x.split('.')[0]))
         
        self.intrincics_color = torch.from_numpy(np.loadtxt(os.path.join(self.root, 'intrinsics', 'intrinsic_color.txt'))).float()
        self.intrincics_depth = torch.from_numpy(np.loadtxt(os.path.join(self.root, 'intrinsics', 'intrinsic_depth.txt'))).float()
        self.extrincics_color = torch.from_numpy(np.loadtxt(os.path.join(self.root, 'intrinsics', 'extrinsic_color.txt'))).float()
        self.extrincics_depth = torch.from_numpy(np.loadtxt(os.path.join(self.root, 'intrinsics', 'extrinsic_depth.txt'))).float()
        
        self.intrincics_color = self.intrincics_color[:3, :3]
        self.intrincics_depth = self.intrincics_depth[:3, :3]
        
        self.idx_to_id = {i: int(f.split('.')[0]) for i, f in enumerate(self.all_color_files)}
        
        
    def __len__(self):
        return len(self.all_color_files)
        
    def __getitem__(self, idx):
        idx = self.idx_to_id[idx]
        
        image = load_image(os.path.join(self.root, 'color', str(idx) + '.jpg'))
        depth = load_depth(os.path.join(self.root, 'depth', str(idx) + '.png')) / 1000
        pose =  torch.from_numpy(np.loadtxt(os.path.join(self.root, 'poses', str(idx) + '.txt'))).float()
        
        image_key = os.path.join(self.root, 'color', str(idx) + '.jpg')
        image_key = os.path.join(*image_key.split('/')[-4:])
        
        depth = torch.nn.functional.interpolate(depth.unsqueeze(0), size=(image.shape[1], image.shape[2]), mode='nearest').squeeze(0).squeeze(0)
        
        return {
            'image': image, # 3x968x1296
            'depth': depth, # 1x480x640 in meters
            'pose': pose, # 3x4
            'K': self.intrincics_color, # 3x3
            'image_id': idx,
            'image_key': image_key,
        }
        
class ScannetPairsScene(torch.utils.data.Dataset):
    defalut_config = {
        "data_path": "./datasets/scannet/scans/scene0000_00",
        "min_overlap_score": 0.2,
        "max_overlap_score": 0.8,
        "mode": "train",
        "resize": 512, #-1
    }
    
    def __init__(self, config={}):
        self.config = config = OmegaConf.merge(OmegaConf.create(self.defalut_config), config)
        self.covis_file = os.path.join(config.data_path, 'covisibility.txt')
        self.covis_pairs = np.loadtxt(self.covis_file, delimiter=',')
        self.covis_pairs = self.covis_pairs[(self.covis_pairs[:, 2] > config.min_overlap_score) & (self.covis_pairs[:, 2] < config.max_overlap_score)]
        
        self.intrincics_color = torch.from_numpy(np.loadtxt(os.path.join(config.data_path, 'intrinsics', 'intrinsic_color.txt'))).float()[0:3, 0:3]
        
    def __len__(self):
        return self.covis_pairs.shape[0]
        
    def __getitem__(self, idx):
        idx0, idx1, covis = self.covis_pairs[idx]
        idx0 = int(idx0)
        idx1 = int(idx1)
        
        img0_key = os.path.join(self.config.data_path, 'color', str(idx0) + '.jpg')
        img1_key = os.path.join(self.config.data_path, 'color', str(idx1) + '.jpg')
        img0_key = os.path.join(*img0_key.split('/')[-4:])
        img1_key = os.path.join(*img1_key.split('/')[-4:])
        
        image0 = load_image(os.path.join(self.config.data_path, 'color', str(idx0) + '.jpg'))
        image1 = load_image(os.path.join(self.config.data_path, 'color', str(idx1) + '.jpg'))

        depth0 = load_depth(os.path.join(self.config.data_path, 'depth', str(idx0) + '.png')) / 1000
        depth1 = load_depth(os.path.join(self.config.data_path, 'depth', str(idx1) + '.png')) / 1000

        # resize depth to image size
        depth0 = torch.nn.functional.interpolate(depth0.unsqueeze(0), size=(image0.shape[1], image0.shape[2]), mode='nearest').squeeze(0).squeeze(0)
        depth1 = torch.nn.functional.interpolate(depth1.unsqueeze(0), size=(image1.shape[1], image1.shape[2]), mode='nearest').squeeze(0).squeeze(0)

        K = self.intrincics_color
        
        if self.config.resize > 0:
            # rescale the images
            image0, scale = resize_short_edge(image0, self.config.resize)
            image1, _ = resize_short_edge(image1, self.config.resize)
            depth0, _ = resize_short_edge(depth0, self.config.resize)
            depth1, _ = resize_short_edge(depth1, self.config.resize)
            
            # crop the images to square            
            image0 = crop_square(image0)
            image1 = crop_square(image1)
            depth0 = crop_square(depth0)
            depth1 = crop_square(depth1)

            # rescale the intrinsics
            K = K * scale
            K[2, 2] = 1
            
        pose0 =  torch.from_numpy(np.loadtxt(os.path.join(self.config.data_path, 'poses', str(idx0) + '.txt'))).float()
        pose1 =  torch.from_numpy(np.loadtxt(os.path.join(self.config.data_path, 'poses', str(idx1) + '.txt'))).float()

        T_0to1 = get_relative_transform(pose0, pose1)
        T_1to0 = get_relative_transform(pose1, pose0)

        return {
            'image0': image0, # 3x968x1296
            'image1': image1, # 3x968x1296
            'depth0': depth0, # 1x480x640
            'depth1': depth1, # 1x480x640
            'pose0': pose0, # 4x4
            'pose1': pose1, # 4x4
            'T_0to1': T_0to1, # 3x4
            'T_1to0': T_1to0, # 3x4
            'K': K, # 3x3
            'K0': K, # 3x3
            'K1': K, # 3x3
            'covis': covis,
            'dataset_name': 'scannet',
            'image0_key': img0_key,
            'image1_key': img1_key,
        }

class ScannetSingleH5(torch.utils.data.Dataset):
    def __init__(self, h5_file, only_images=False, features_cache=None, dino_cache=None):
        super().__init__()

        self.h5_file = h5_file
        self.only_images = only_images
        self.features_cache = features_cache
        self.dino_cache = dino_cache
        if only_images:
            cached_keys_fname = h5_file + '.keys.images'
            if os.path.exists(cached_keys_fname):
                with open(cached_keys_fname, 'r') as f:
                    self.keys = f.readlines()
                    self.keys = [k.strip() for k in self.keys]
            else:
                with h5py.File(h5_file, 'r') as h5:
                    self.keys = list(h5['images'].keys())
                with open(cached_keys_fname, 'w') as f:
                    for key in self.keys:
                        f.write(f"{key}\n")
        else:
            cached_keys_fname = h5_file + '.keys.pairs'
            if os.path.exists(cached_keys_fname):
                with open(cached_keys_fname, 'r') as f:
                    self.keys = f.readlines()
                    self.keys = [k.strip() for k in self.keys]
            else:
                with h5py.File(h5_file, 'r') as h5:
                    self.keys = list(h5['T_0to1'].keys())
                with open(cached_keys_fname, 'w') as f:
                    for key in self.keys:
                        f.write(f"{key}\n")
        
    def __len__(self):
        return len(self.keys)
    
    def __getitem__(self, idx):
        if self.only_images:
            with h5py.File(self.h5_file, 'r') as h5:
                image = h5['images'][self.keys[idx]][...]
                return {
                    'image': torch.tensor(image),
                    'image_key': self.keys[idx],
                }
        
        key0, key1 = self.keys[idx].split(':')

        with h5py.File(self.h5_file, 'r') as h5:
            T_0to1 = h5['T_0to1'][self.keys[idx]][...]
            T_1to0 = h5['T_1to0'][self.keys[idx]][...]
            covis = h5['covis'][self.keys[idx]][...]

            image0 = h5['images'][key0][...]
            image1 = h5['images'][key1][...]
            depth0 = h5['depth'][key0][...]
            depth1 = h5['depth'][key1][...]

            K0 = h5['K'][key0][...]
            K1 = h5['K'][key1][...]
            
        return_data = {
            'image0': torch.tensor(image0),
            'image1': torch.tensor(image1),
            'depth0': torch.tensor(depth0),
            'depth1': torch.tensor(depth1),
            'T_0to1': torch.tensor(T_0to1),
            'T_1to0': torch.tensor(T_1to0),
            'K0': torch.tensor(K0),
            'K1': torch.tensor(K1),
            'covis': torch.tensor(covis),
            'dataset': 'scannet',
        }
        
        if self.features_cache is not None:
            with h5py.File(self.features_cache, 'r') as h5:
                for cache_key in ['keypoints', 'descriptors']:
                    return_data[f'{cache_key}0'] = torch.tensor(h5[cache_key][key0][...]).float()
                    return_data[f'{cache_key}1'] = torch.tensor(h5[cache_key][key1][...]).float()
                    
        if self.dino_cache is not None:
            with h5py.File(self.dino_cache, 'r') as h5:
                dino_map0 = torch.tensor(h5['features'][key0][...])
                dino_map1 = torch.tensor(h5['features'][key1][...])
                keypoints0 = return_data['keypoints0']
                keypoints1 = return_data['keypoints1']
                
                # sample_features_dino
                semanitc_features0 = sample_features_dino(keypoints0.unsqueeze(0), dino_map0.unsqueeze(0))[0]
                semanitc_features1 = sample_features_dino(keypoints1.unsqueeze(0), dino_map1.unsqueeze(0))[0]
                
                return_data['semantic_features0'] = semanitc_features0
                return_data['semantic_features1'] = semanitc_features1
                
        # normalize size to be fixed at 32 chars
        image0_key = key0.ljust(40, '#')
        image1_key = key1.ljust(40, '#')
        return_data['image0_key'] = image0_key
        return_data['image1_key'] = image1_key
        
        return return_data

class ScannetH5(torch.utils.data.Dataset):
    def __init__(self, root, split, max_samples=-1, only_images=False, features_cache=None, dino_cache=None):
        super().__init__()
        assert split in ['train', 'val', 'test']
        assert os.path.exists(root)
        self.root = root
        self.split = split
        self.max_samples = max_samples
        self.h5_files = sorted(glob(os.path.join(root, split, '*.h5')))
        assert len(self.h5_files) > 0, f"No h5 files found in {root}/{split}"
        
        if features_cache is not None:
            h5_files_features_cache = sorted(glob(os.path.join(root, 'features', features_cache, split, '*.h5')))
            if len(h5_files_features_cache) == 0:
                raise ValueError(f"No features cache found in {root}/features/{features_cache}/{split}")
            if len(h5_files_features_cache) != len(self.h5_files):
                raise ValueError(f"Features cache and h5 files do not match in length")
        else:
            h5_files_features_cache = [None] * len(self.h5_files)
            
        if dino_cache is not None:
            h5_file_dino_cache = sorted(glob(os.path.join(root, 'features', dino_cache, split, '*.h5')))
            if len(h5_file_dino_cache) == 0:
                raise ValueError(f"No dino cache found in {root}/features/{dino_cache}/{split}")
            if len(h5_file_dino_cache) != len(self.h5_files):
                raise ValueError(f"Dino cache and h5 files do not match in length")
        else:
            h5_file_dino_cache = [None] * len(self.h5_files)
        
        self.dataset = torch.utils.data.ConcatDataset([
            ScannetSingleH5(h5_file, only_images=only_images, features_cache=features_cache, dino_cache=dino_cache)
            for (h5_file, features_cache, dino_cache) 
            in tqdm(zip(self.h5_files, h5_files_features_cache, h5_file_dino_cache), total=len(self.h5_files), desc=f"Loading {split} h5 files")
        ])

    def __len__(self):
        if self.max_samples > 0:
            return min(len(self.dataset), self.max_samples)
        return len(self.dataset)
    
    def __getitem__(self, idx):
        return self.dataset[idx]

def makeScannetDataset(scannet_scenes_folder, min_overlap_score, max_overlap_score, mode, max_scenes=-1, concat=True):
    assert mode in ['train', 'val', 'test']
    this_folder = os.path.dirname(os.path.abspath(__file__))
    split = os.path.join(this_folder, 'scannetv1_' + mode + '.txt')
    with open(split, 'r') as f:
        scenes = f.readlines()
    scenes = sorted([s.strip() for s in scenes])
    if max_scenes > 0:
        scenes = scenes[:max_scenes]
    if max_scenes == 1:
        scenes = ['scene0000_00']
    
    if concat:
        concat_dataset = torch.utils.data.ConcatDataset([
            ScannetPairsScene({
                "data_path": os.path.join(scannet_scenes_folder, s),
                "min_overlap_score": min_overlap_score,
                "max_overlap_score": max_overlap_score,
                "mode": mode,
            }) for s in scenes
        ])
    else:
        concat_dataset = {
            s: ScannetPairsScene({
                "data_path": os.path.join(scannet_scenes_folder, s),
                "min_overlap_score": min_overlap_score,
                "max_overlap_score": max_overlap_score,
                "mode": mode,
            }) for s in scenes
        }
        # scene0011_00, scene0011_01, scene0011_02 should be in the same dataset
        joint = {}
        for s, dataset in concat_dataset.items():
            if s.split('_')[0] not in joint:
                joint[s.split('_')[0]] = []
            joint[s.split('_')[0]].append(dataset)

        for s, datasets in joint.items():
            joint[s] = torch.utils.data.ConcatDataset(datasets)
            
        concat_dataset = joint
        
    return concat_dataset     

def calculate_covisibility(scene_folder):

    scannet_images = ScannetImages({
        'data_path': scene_folder,
    })
    
    covis_file = open( os.path.join(scannet_images.root, 'covisibility.txt'), 'w')

    norm_grid = torch.stack(torch.meshgrid(torch.linspace(0, 1, 32), torch.linspace(0, 1, 32)), dim=-1).reshape(-1, 2)

    for source_idx in tqdm(range(len(scannet_images)), position=0, leave=True):
        data_source = add_batch_dim(scannet_images[source_idx])
        
        for target_idx in range(source_idx + 1, len(scannet_images)):            
            data_target = add_batch_dim(scannet_images[target_idx])
            kps0 = (norm_grid * torch.tensor([data_source['image'].shape[3]-1, data_source['image'].shape[2]-1]).float()).unsqueeze(0) # BxNx2
            T_0to1 = get_relative_transform(data_source['pose'], data_target['pose'])
            valid_mask, w_kpts0 = warp_kpts(kps0, data_source['depth'], data_target['depth'], T_0to1, data_source['K'], data_target['K'])
            covis = valid_mask.sum().item() / valid_mask.shape[1]
            covis_file.write(f"{data_source['image_id']},{data_target['image_id']},{covis}\n")
            covis_file.flush()

    covis_file.close()
    


# Copyright 2019-present NAVER Corp.
# CC BY-NC-SA 3.0
# Available only for non-commercial use

import pdb
import numpy as np
import torch
import torch.nn as nn


class APLoss (nn.Module):
    """ differentiable AP loss, through quantization.
        
        Input: (N, M)   values in [min, max]
        label: (N, M)   values in {0, 1}
        
        Returns: list of query AP (for each n in {1..N})
                 Note: typically, you want to minimize 1 - mean(AP)
    """
    def __init__(self, nq=25, min=0, max=1, euc=False):
        nn.Module.__init__(self)
        assert isinstance(nq, int) and 2 <= nq <= 100
        self.nq = nq
        self.min = min
        self.max = max
        self.euc = euc
        gap = max - min
        assert gap > 0
        
        # init quantizer = non-learnable (fixed) convolution
        self.quantizer = q = nn.Conv1d(1, 2*nq, kernel_size=1, bias=True)
        a = (nq-1) / gap
        #1st half = lines passing to (min+x,1) and (min+x+1/a,0) with x = {nq-1..0}*gap/(nq-1)
        q.weight.data[:nq] = -a
        q.bias.data[:nq] = torch.from_numpy(a*min + np.arange(nq, 0, -1)) # b = 1 + a*(min+x)
        #2nd half = lines passing to (min+x,1) and (min+x-1/a,0) with x = {nq-1..0}*gap/(nq-1)
        q.weight.data[nq:] = a
        q.bias.data[nq:] = torch.from_numpy(np.arange(2-nq, 2, 1) - a*min) # b = 1 - a*(min+x)
        # first and last one are special: just horizontal straight line
        q.weight.data[0] = q.weight.data[-1] = 0
        q.bias.data[0] = q.bias.data[-1] = 1

    def compute_AP(self, x, label):
        N, M = x.shape
        if self.euc:  # euclidean distance in same range than similarities
            x = 1 - torch.sqrt(2.001 - 2*x)

        # quantize all predictions
        q = self.quantizer(x.unsqueeze(1))
        q = torch.min(q[:,:self.nq], q[:,self.nq:]).clamp(min=0) # N x Q x M

        nbs = q.sum(dim=-1) # number of samples  N x Q = c
        rec = (q * label.view(N,1,M).float()).sum(dim=-1) # nb of correct samples = c+ N x Q
        prec = rec.cumsum(dim=-1) / (1e-16 + nbs.cumsum(dim=-1)) # precision
        rec /= rec.sum(dim=-1).unsqueeze(1) # norm in [0,1]

        ap = (prec * rec).sum(dim=-1) # per-image AP
        return ap

    def forward(self, x, label):
        assert x.shape == label.shape # N x M
        return self.compute_AP(x, label)








import torch, typing
import numpy as np
# from torch_dimcheck import dimchecked
# from ...submodules.torch_dimcheck import dimchecked

from ..common.structs import NpArray, Features, MatchedPairs
from ..geom import distance_matrix

class CycleMatcher:
    # @dimchecked
    def match_features(self, feat_1, feat_2):
        dist_m = distance_matrix(feat_1, feat_2)

        if dist_m.shape[0] == 0 or dist_m.shape[1] == 0:
            msg = '''
            Feature matching failed because one image has 0 detected features.
            This likely means that the algorithm has converged to a local
            optimum of detecting no features at all (0 reward). This can arise
            when lambda_fp and lambda_kp penalties are too high. Please check
            that your penalty annealing scheme is sound. It can also be that
            you are using a too low value of --warmup or --chunk-size
            '''
            raise RuntimeError(msg)

        n_amin = torch.argmin(dist_m, dim=1)
        m_amin = torch.argmin(dist_m, dim=0)

        # nearest neighbor's nearest neighbor
        nnnn = m_amin[n_amin]

        # we have a cycle consistent match for each `i` such that
        # nnnn[i] == i. We create an auxiliary array to check for that
        n_ix = torch.arange(dist_m.shape[0], device=dist_m.device)
        mask = nnnn == n_ix

        # Now `mask` is a binary mask and n_amin[mask] is an index array.
        # We use nonzero to turn `n_amin[mask]` into an index array and return
        return torch.stack([
            torch.nonzero(mask, as_tuple=False)[:, 0],
            n_amin[mask],
        ], dim=0)

    def match_pairwise(
        self,
        features: NpArray[Features], # [N_scenes, N_per_scene]
    ): # -> [N_scenes, (N_per_scene choose 2)]
        N_scenes, N_per_scene = features.shape
        N_combinations        = N_per_scene * (N_per_scene - 1) // 2

        matched_pairs = np.zeros((N_scenes, N_combinations), dtype=object)
        for i_scene, scene_f in enumerate(features):
            i_decision = 0

            for i in range(N_per_scene):
                features1 = scene_f[i]
                for j in range(i+1, N_per_scene):
                    features2 = scene_f[j]

                    matched_pairs[i_scene, i_decision] = MatchedPairs(
                        features1.kp,
                        features2.kp,
                        self.match_features(features1.desc, features2.desc),
                    )
                    
                    i_decision += 1

        return matched_pairs



import typing, torch, math
import numpy as np
import torch.nn.functional as F

# from torch_dimcheck import dimchecked
# from ...submodules.torch_dimcheck import dimchecked

RAD_TO_DEG = 57.29

class PoseError(typing.NamedTuple):
    Δ_θ    : float
    Δ_T    : float

    def __repr__(self):
        return f'<PoseError Δ_θ={self.Δ_θ:.3f}, Δ_T={self.Δ_T:.3f}>'

    def to_dict(self):
        return {
            'Δ_θ'    : self.Δ_θ,
            'Δ_T'    : self.Δ_T,
        }

class Pose(typing.NamedTuple):
    R: torch.Tensor
    T: torch.Tensor

    def __str__(self):
        return f'\nR=\n{self.R}\nT=\n{self.T}'

    def to(self, *args, **kwargs):
        return Pose(
            self.R.to(*args, **kwargs),
            self.T.to(*args, **kwargs),
        )
    
    @property
    def R_4x4(self):
        I = torch.eye(4)
        I[:3, :3] = self.R

        return I

    @property
    def T_4x4(self):
        I = torch.eye(4)
        I[:3, 3] = self.T

        return I
        
    def TR_4x4(self):
        return self.T_4x4 @ self.R_4x4

    @classmethod
    def from_poselike(cls, poselike):
        return cls(poselike.R, poselike.T)

    @staticmethod
    def relative(p1, p2, normed=False):
        assert isinstance(p1, Pose)
        assert isinstance(p2, Pose)

        ΔR = p2.R @ p1.R.T
        ΔT = p2.T - ΔR @ p1.T

        if normed:
            ΔT = ΔT / torch.norm(ΔT, p=2)

        return Pose(ΔR, ΔT)

    @staticmethod
    def error(p1, p2):
        assert isinstance(p1, Pose)
        assert isinstance(p2, Pose)

        R_ab       = p1.R.T @ p2.R
        half_trace = (torch.trace(R_ab) - 1) / 2
        θ_deg      = RAD_TO_DEG * math.acos(torch.clamp(half_trace, -1., 1.).item()) / 2

        cos_sim = F.cosine_similarity(
            p1.T.unsqueeze(0),
            p2.T.unsqueeze(0)
        ).squeeze(0)
        Δ_T = RAD_TO_DEG * math.acos(torch.clamp(cos_sim, -1., 1.).item())
            
        return PoseError(θ_deg, Δ_T)
        
def _normalized_cosine_error(v1, v2):
    EPS = 1e-15

    def normalize(v):
        v = v.to(torch.float64)
        return v / (torch.norm(v) + EPS)

    v1 = normalize(v1)
    v2 = normalize(v2)

    cos = max(EPS, 1. - torch.dot(v1, v2).pow(2).item())
    err = math.acos(math.sqrt(1. - cos))

    return err

# @dimchecked
def matrix_to_quaternion(M):
    '''
    adapted from
    https://github.com/vcg-uvic/sfm_benchmark/blob/2b28c76635f754cbc32f30571adf80f3eba13f4c/utils/eval_helper.py#L170
    '''

    #TODO: this appears unused
    m00 = M[0, 0].item()
    m01 = M[0, 1].item()
    m02 = M[0, 2].item()
    m10 = M[1, 0].item()
    m11 = M[1, 1].item()
    m12 = M[1, 2].item()
    m20 = M[2, 0].item()
    m21 = M[2, 1].item()
    m22 = M[2, 2].item()

    # symmetric matrix K
    K = np.array([
        [m00 - m11 - m22,       0.0,                         0.0,             0.0],
        [      m01 + m10,       m11 - m00 - m22,             0.0,             0.0],
        [      m02 + m20,             m12 + m21, m22 - m00 - m11,             0.0],
        [      m21 - m12,             m02 - m20,       m10 - m01, m00 + m11 + m22],
    ])
    K /= 3.0
    # quaternion is eigenvector of K that corresponds to largest eigenvalue
    w, V = np.linalg.eigh(K)
    q = V[[3, 0, 1, 2], np.argmax(w)]

    q = torch.from_numpy(q)

    if q[0] > 0.0:
        return q
    else:
        return -q



import pytest
import torch

from easy_local_features.feature.baseline_desc_reasoning import Desc_Reasoning_baseline
from easy_local_features.utils import io, ops


@pytest.mark.timeout(300)
def test_desc_reasoning_detect_and_match():
    # Use smallest weights to reduce download time
    try:
        method = Desc_Reasoning_baseline(
            {
                "pretrained": "xfeat-3_layers",
                "device": "cpu",
            }
        )
    except Exception as e:
        pytest.skip(f"Skipping due to environment/network issue: {e}")

    img0 = io.fromPath("tests/assets/megadepth0.jpg")
    img1 = io.fromPath("tests/assets/megadepth1.jpg")
    img0 = ops.resize_short_edge(img0, 320)[0]
    img1 = ops.resize_short_edge(img1, 320)[0]

    kpts0, desc0 = method.detectAndCompute(img0)
    assert isinstance(kpts0, torch.Tensor)
    assert isinstance(desc0, torch.Tensor)
    assert kpts0.ndim == 3 and kpts0.shape[-1] == 2  # [B, N, 2]
    assert desc0.ndim == 4 and desc0.shape[-1] == 2  # [B, N, D, 2]

    res = method.match(img0, img1)
    mk0, mk1 = res["mkpts0"], res["mkpts1"]
    assert isinstance(mk0, torch.Tensor)
    assert isinstance(mk1, torch.Tensor)
    assert mk0.ndim == 2 and mk0.shape == mk1.shape
